import numpy as np
from ._utils import _array_input


class covFunction:
    """
    Base class for covariance functions (kernels) used for nodes and weights in
    the GPRN.
    """
    def __init__(self, *args):
        self.pars = np.array(args, dtype=float)

    def __call__(self, r, t1=None, t2=None):
        raise NotImplementedError

    def _dkdxidj(self, r):
        raise NotImplementedError

    def __repr__(self):
        if hasattr(self, '_param_names'):
            pars = ', '.join(
                [f'{p}={v}' for p, v in zip(self._param_names, self.pars)])
        else:
            pars = ', '.join(map(str, self.pars))
        return f"{self.__class__.__name__}({pars})"

    def get_parameters(self):
        return self.pars

    @_array_input
    def set_parameters(self, p):
        msg = f'too few parameters for kernel {self.__class__.__name__}'
        assert len(p) >= self.pars.size, msg
        if len(p) > self.pars.size:
            p = list(p)
            self.pars = np.array(p[:self.pars.size], dtype=float)
            for _ in range(self.pars.size):
                p.pop(0)
            return np.array(p)
        else:
            self.pars = p

    def __add__(self, b):
        return Sum(self, b)

    def __radd__(self, b):
        return self.__add__(b)

    def __mul__(self, b):
        return Multiplication(self, b)

    def __rmul__(self, b):
        return self.__mul__(b)


class _operator(covFunction):
    """ To allow operations between two kernels """
    def __init__(self, k1, k2):
        self.k1 = k1
        self.k2 = k2
        self.kerneltype = 'complex'
        self.pars = np.r_[self.k1.pars, self.k2.pars]


class Sum(_operator):
    """ Sum of two covariance functions """
    def __call__(self, r):
        return self.k1(r) + self.k2(r)

    def __repr__(self):
        return "{0} + {1}".format(self.k1, self.k2)


class Multiplication(_operator):
    """ Product of two covariance functions """
    def __call__(self, r):
        return self.k1(r) * self.k2(r)

    def __repr__(self):
        return "{0} * {1}".format(self.k1, self.k2)


class _unary_operator(covFunction):
    """ To allow operations for one kernel """
    def __init__(self, k):
        msg = f'kernel {k} is not twice differentiable'
        if not hasattr(k, '_twice_differentiable') or not k._twice_differentiable:
            raise ValueError(msg)

        self.k = k
        self.kerneltype = 'complex_unary'
        self.pars = self.k.pars

        self._param_names = self.k._param_names
        self._tag = 'd' + self.k._tag


class Derivative(_unary_operator):
    def __call__(self, r):
        return self.k._dkdxidj(r)

    def __repr__(self):
        self.k.pars = self.pars
        return "d {0}".format(self.k)


class Constant(covFunction):
    """
    This kernel returns the square of its constant argument c
    $$
      K_{ij} = c^2
    $$

    Args:
        c: Constant value
    """
    _param_names = 'c',
    _tag = 'C'

    def __init__(self, c: float):
        super(Constant, self).__init__(c)

    def __call__(self, r):
        c = self.pars[0]
        return np.full_like(r, c**2)


class WhiteNoise(covFunction):
    """
    White noise (diagonal) kernel
    $$
      K_{ij} = w^2 \, \delta_{ij}
    $$

    Args:
        w: White noise amplitude
    """
    _param_names = 'wn',
    _tag = 'WN'

    def __init__(self, w:float):
        super(WhiteNoise, self).__init__(w)

    def __call__(self, r):
        w = self.pars[0]
        if r.ndim == 2 and r[0, :].shape == r[:, 0].shape:
            return w**2 * np.diag(np.diag(np.ones_like(r)))
        return np.full_like(r, w**2)


class SquaredExponential(covFunction):
    r"""
    Squared Exponential kernel, also known as radial basis function
    $$
      K_{ij} = \theta^2 \, \exp \left[ - \frac{(t_i - t_j)^2}{2 \ell^2} \right]
    $$

    Args:
        theta: Amplitude
        ell: Length-scale
    """
    _param_names = 'theta', 'ell'
    _tag = 'SE'
    _twice_differentiable = True

    def __init__(self, theta: float, ell: float):
        super(SquaredExponential, self).__init__(theta, ell)

    def __call__(self, r):
        return self.pars[0]**2 * np.exp(-0.5 * r**2 / self.pars[1]**2)

    def _dkdxi(self, r):
        θ, ell = self.pars
        return θ**2 * (-r) * np.exp(-0.5 * (-r)**2 / ell**2) / ell**2

    def _dkdxj(self, r):
        # covariance between an observation at xi
        # and a derivative observation at xj
        θ, ell = self.pars
        return θ**2 * r * np.exp(-0.5 * r**2 / ell**2) / ell**2

    def _dkdxidj(self, r):
        term1 = self.pars[0]**2 / self.pars[1]**4
        term2 = self.pars[1]**2 - r**2
        return term1 * term2 * np.exp(-0.5 * r**2 / self.pars[1]**2)


class Periodic(covFunction):
    r"""
    Periodic kernel, also known as the exponential sine squared
    $$
      K_{ij} = \theta^2 \, \exp \left[ - \frac{2 \sin^2 \left( \frac{\pi (t_i - t_j)}{P} \right)}{\ell^2} \right]
    $$

    Args:
        theta: Amplitude
        P: Period
        ell: Lenght scale

    Note: Parameterization
        Note that the periodic kernel is sometimes parameterized differently,
        namely using $\Gamma = 2/\ell^2$.  
    """
    _param_names = 'theta', 'P', 'ell'
    _tag = 'P'
    _twice_differentiable = True

    def __init__(self, theta: float, P: float, ell: float):
        super(Periodic, self).__init__(theta, P, ell)

    def __call__(self, r):
        θ, P, ell = self.pars
        return θ**2 * np.exp(-2 * np.sin(np.pi * np.abs(r) / P)**2 / ell**2)

    def _dkdxidj(self, r):
        θ, P, ell = self.pars
        rP = np.pi * r / P
        term1 = 4 * np.pi**2 * θ**2
        term2 = ell**2 * np.cos(2 * rP) - 4 * np.sin(rP)**2 * np.cos(rP)**2
        term3 = np.exp(-2 * np.sin(rP)**2 / ell**2)
        return term1 * term2 * term3


class QuasiPeriodic(covFunction):
    r"""
    This kernel is the product between the periodic kernel and the squared
    exponential kernel, and is commonly known as the quasi-periodic kernel
    $$
      K_{ij} = \theta^2 \,
                  \exp \left[ - \frac{(t_i - t_j)^2}{2 \ell_e^2} - \frac{2 \sin^2 \left( \frac{\pi (t_i - t_j)}{P} \right)}{\ell_p^2} \right]
    $$

    Args:
        theta: Amplitude
        elle: Evolutionary length scale
        P: Kernel periodicity
        ellp: Length scale of the periodic component

    Info:
        The `QuasiPeriodic` kernel is implemented on its own for convenience,
        but it is exactly equivalent to the product of a `SquaredExponential`
        and a `Periodic` kernel with the same parameters.
    """
    _param_names = 'theta', 'le', 'P', 'lp'
    _tag = 'QP'
    _twice_differentiable = True

    def __init__(self, theta: float, elle: float, P: float, ellp: float):
        super(QuasiPeriodic, self).__init__(theta, elle, P, ellp)

    def __call__(self, r):
        θ, elle, P, ellp = self.pars
        term1 = -2 * np.sin(np.pi * np.abs(r) / P)**2 / ellp**2
        term2 = r**2 / (2 * elle**2)
        return θ**2 * np.exp(term1 - term2)

    def _dkdxidj(self, r):
        θ, elle, P, ellp = self.pars
        term1 = 2 * θ**2 / (P**2 * ellp**4 * elle**4)
        term2 = P**2 * ellp**4 * elle**2 - \
                2 * P**2 * ellp**4 * r**2 - \
                4 * np.pi * P * ellp**2 * elle**2 * r * np.sin(2 * np.pi * r / P) + \
                2 * np.pi**2 * ellp**2 * elle**4 * np.cos(2 * np.pi * r / P) - \
                8 * np.pi**2 * elle**4 * np.sin(np.pi * r / P)**2 * np.cos(np.pi * r / P)**2
        term3 = np.exp(-(ellp**2 * r**2 + 2 * elle**2 * np.sin(np.pi * r / P)**2) / (ellp**2 * elle**2))
        return term1 * term2 * term3


class RationalQuadratic(covFunction):
    """
    The rational quadratic kernel
    $$
    $$

    Args:
        theta: Amplitude of the kernel
        alpha: Amplitude of large and small scale variations
        ell: Characteristic lenght scale to define the kernel "smoothness"
    """
    _param_names = 'theta', 'alpha', 'ell'
    _tag = 'RQ'

    def __init__(self, theta: float, alpha: float, ell: float):
        super(RationalQuadratic, self).__init__(theta, alpha, ell)

    def __call__(self, r):
        θ, α, ell = self.pars
        return θ**2 * (1 + 0.5 * r**2 / (α * ell**2))**(-α)


class RQP(covFunction):
    """
    Product between the periodic kernel and the rational quadratic kernel that
    we call RQP kernel.

    Args:
        theta: Amplitude
        alpha: Alpha of the rational quadratic kernel
        elle: Aperiodic length scale
        P: Periodic repetitions of the kernel
        ellp: Periodic length scale
    """
    _param_names = 'theta', 'alpha', 'elle', 'ellp', 'P'
    _tag = 'RQP'

    def __init__(self, theta: float, alpha: float, elle: float, P: float,
                 ellp: float):
        super(RQP, self).__init__(theta, alpha, elle, P, ellp)

    def __call__(self, r):
        θ, α, ℓe, P, ℓp = self.pars
        return θ**2 * np.exp(-2 * np.sin(np.pi * np.abs(r) / P)**2 /
                             ℓp**2) * (1 + r**2 / (2 * α * ℓe**2))**(-α)


class COSINE(covFunction):
    """
    the cosine kernel

    Args:
        theta: Amplitude
        P: Period
    """
    _param_names = 'theta', 'P'
    _tag = 'COS'

    def __init__(self, theta: float, P: float):
        super(COSINE, self).__init__(theta, P)

    def __call__(self, r):
        return self.pars[0]**2 * np.cos(2 * np.pi * np.abs(r) / self.pars[1])


class Exponential(covFunction):
    r"""
    The exponential kernel
    $$
      K_{ij} = \theta^2 \, \exp \left( - \frac{|t_i - t_j|}{\ell} \right)
    $$

    Args:
        theta: Amplitude
        ell: Characteristic lenght scale
    """
    _param_names = 'theta', 'ell'
    _tag = 'EXP'

    def __init__(self, theta: float, ell: float):
        super(Exponential, self).__init__(theta, ell)

    def __call__(self, r):
        return self.pars[0]**2 * np.exp(-np.abs(r) / self.pars[1])


class Matern32(covFunction):
    """
    the Matern 3/2 kernel. This kernel arise when setting
    v=3/2 in the matern family of kernels

    Args:
        theta: Amplitude
        ell: Characteristic lenght scale
    """
    _param_names = 'theta', 'ell'
    _tag = 'M32'

    def __init__(self, theta: float, ell: float):
        super(Matern32, self).__init__(theta, ell)

    def __call__(self, r):
        return self.pars[0]**2 * (
            1.0 + np.sqrt(3.0) * np.abs(r) / self.pars[1]) * np.exp(
                -np.sqrt(3.0) * np.abs(r) / self.pars[1])


#### Matern 5/2 ################################################################
class Matern52(covFunction):
    """
    the Matern 5/2 kernel. This kernel arise when setting v=5/2
    in the matern family of kernels

    Args:
        theta: Amplitude
        ell: Characteristic lenght scale
    """
    _param_names = 'theta', 'ell'
    _tag = 'M52'

    def __init__(self, theta: float, ell: float):
        super(Matern52, self).__init__(theta, ell)

    def __call__(self, r):
        return self.pars[0]**2 * (
            1.0 +
            (3 * np.sqrt(5) * self.pars[1] * np.abs(r) + 5 * np.abs(r)**2) /
            (3 * self.pars[1]**2)) * np.exp(
                -np.sqrt(5.0) * np.abs(r) / self.pars[1])


class Linear(covFunction):
    """
    the Linear kernel

    Args:
        c: Constant
    """
    def __init__(self, c):
        super(Linear, self).__init__(c)
        self.tag = 'LIN'
        self.c = c

    def __call__(self, r, t1, t2):
        return  (t1 - self.pars[0]) * (t2 - self.pars[0])


class GammaExp(covFunction):
    """
    the gamma-exponential kernel

    Args:
        theta: Amplitude
        gamma: Shape parameter ( 0 < gamma <= 2)
        l: Lenght scale
    """
    def __init__(self, theta, gamma, l):
        super(GammaExp, self).__init__(theta, gamma, l)
        self.tag = 'GammaExp'
        self.theta = theta
        self.gamma = gamma
        self.l = l

    def __call__(self, r):
        return self.pars[0]**2 *np.exp(-(np.abs(r)/self.pars[2])**self.pars[1])


class Polynomial(covFunction):
    """
    the polynomial kernel

    Args:
        theta: Amplitude ???
        a: Real value > 0
        b: Real value >= 0
        c: Integer value
        wn: White noise amplitude
    """
    def __init__(self, theta, a, b, c):
        super(Polynomial, self).__init__(theta, a, b, c)
        self.tag = 'POLY'
        self.theta = theta
        self.a = a
        self.b = b
        self.c = c

    def __call__(self, r, t1, t2):
        return (self.pars[1] * t1 * t2 + self.pars[2])**self.pars[3]


##### Piecewise ################################################################
class Piecewise(covFunction):
    """
    WARNING: EXPERIMENTAL KERNEL
    
    Parameters
    ----------
    """
    def __init__(self, eta):
        super(Piecewise, self).__init__(eta)
        self.eta = eta
        self.type = 'unknown'
    def __call__(self, r):
        r = r/(0.5*self.pars[0])
        piecewise = (3*np.abs(r) +1) * (1 - np.abs(r))**3
        piecewise = np.where(np.abs(r)>1, 0, piecewise)
        return piecewise


### END



"""
Auxiliary Gaussian processes functions
"""
import numpy as np
from .covfunc import covFunction as kernels
from scipy.linalg import cho_factor, cho_solve

##### Gaussian processes #######################################################
class GP(object):
    """ 
    Class to create our Gaussian process.
    
    Parameters
    ----------
    kernel: func
        Covariance funtion
    means: func
        Mean function 
    time: array
        Time array
    y: array
        Measurements array
    yerr: array
        Measurements errors array
    """
    def __init__(self, time, y, yerr=None):
        self.time = time            #time
        self.y = y                  #measurements
        if yerr is None:
            self.yerr = 1e-12 * np.identity(self.time.size)
        else:
            self.yerr = np.array(yerr)        #measurements errors
        self.yerr2 = self.yerr**2
        
    def _kernel_pars(self, kernel):
        """ Returns a kernel parameters """
        return kernel.pars

    def _kernel_matrix(self, kernel, time):
        """ Returns a cov matrix when evaluating a given kernel at inputs time """
        r = time[:, None] - time[None, :]
        K = kernel(r) + 1.25e-12*np.diag(np.diag(np.ones_like(r)))
        return K

    def _predict_kernel_matrix(self, kernel, time):
        """ To be used in prediction() """
        r = time[:, None] - self.time[None, :]
        K = kernel(r)
        return K

    def new_kernel(self, kernel, new_pars):
        """
        Updates the parameters of a kernel
        
        Parameters
        ----------
        kernel: func
            Original kernel
        new_pars: list
            New hyperparameters
        
        Returns
        -------
        new_k: func
            Updated kernel
        """
        #if we are working with the sum of kernels
        if isinstance(kernel, kernels.Sum):
            k1_params = []
            for i, j in enumerate(kernel.k1.pars):
                k1_params.append(new_pars[i])
            k2_params = []
            for i, j in enumerate(kernel.k2.pars):
                k2_params.append(new_pars[len(kernel.k1.pars)+i])
            new_k1 = type(kernel.k1)(*k1_params)
            new_k2 = type(kernel.k2)(*k2_params)
            return new_k1+new_k2
        #if we are working with the product of kernels
        elif isinstance(kernel, kernels.Multiplication):
            k1_params = []
            for i, _ in enumerate(kernel.k1.pars):
                k1_params.append(new_pars[i])
            k2_params = []
            for j, _ in enumerate(kernel.k2.pars):
                k2_params.append(new_pars[len(kernel.k1.pars)+j])
            new_k = type(kernel.k1)(*k1_params) * type(kernel.k1)(*k2_params)
            return new_k
        #if we are working with a "single" kernel
        else:
            new_k = type(kernel)(*new_pars)
            return new_k


    def prediction(self, kernel, time, m, v):
        """ 
        Conditional predictive distribution of the Gaussian process
        
        Parameters
        ----------
        kernel: func
            Covariance function
        time: array
            Time array

        Returns
        -------
        y_mean: array
            Mean vector
        y_var: array
            Variance vector
        """
        cov = self._kernel_matrix(kernel, self.time) + np.diag(v) #K
        L1 = cho_factor(cov)
        sol = cho_solve(L1, m)
        #Kstar
        Kstar = self._predict_kernel_matrix(kernel, time)
        #Kstarstar
        Kstarstar =  self._kernel_matrix(kernel, time)
        y_mean = np.dot(Kstar, sol) #mean
        kstarT_k_kstar = []
        for i, _ in enumerate(time):
            kstarT_k_kstar.append(np.dot(Kstar, cho_solve(L1, Kstar[i,:])))
        y_cov = Kstarstar - kstarT_k_kstar
        y_var = np.diag(y_cov) #variance
        return y_mean, y_var

### END



import os, sys
import numpy as np
import matplotlib.pyplot as plt


def equal_y_axis(ax):
    ylim = ax.get_ylim()
    m = np.max(np.abs(ylim))
    ax.set_ylim(-m, m)


def plot_prediction(gprn, nn=1000, tstar=None, over=0.2, title=None):
    if tstar is None:
        mi, ma = gprn.time.min(), gprn.time.max()
        tptp = gprn.time.ptp()
        tstar = np.linspace(mi - over * tptp, ma + over * tptp, nn)

    a, v = gprn._Prediction()
    aa, vv, bb = gprn._Prediction(tstar=tstar, separate=True)
    ss = np.sqrt(vv)

    if gprn.p == 1:
        layout = [
            ['pred1', 'd1'],
            ['resid1', 'node'],
        ]
    elif gprn.p == 2:
        layout = [
            ['pred1', 'd1'],
            ['pred1', 'd1'],
            ['resid1', 'node'],
            ['pred2', 'node'],
            ['pred2', 'd2'],
            ['resid2', 'd2'],
        ]
    elif gprn.p == 3:
        layout = [
            ['node', 'node'],
            ['node', 'node'],
            ['pred1', 'd1'], ['pred1', 'd1'],# ['pred1', 'd1'],
            ['resid1', 'd1'],
            ['pred2', 'd2'], ['pred2', 'd2'],# ['pred2', 'd2'],
            ['resid2', 'd2'],
            ['pred3', 'd3'], ['pred3', 'd3'],# ['pred3', 'd3'],
            ['resid3', 'd3'],
        ]

    fig = plt.figure(constrained_layout=False, figsize=(10, 4 * gprn.p))
    if title is not None:
        fig.suptitle(title)

    axs = fig.subplot_mosaic(layout)

    means = []
    for i in range(gprn.p):
        axs[f'pred{i+1}'].set(xlabel='', ylabel=f'y{i+1}')
        axs[f'pred{i+1}'].errorbar(gprn.time, gprn.y[i], gprn.yerr[i],
                                   fmt='ok', ms=2)

        pred = aa[:, i].T
        std = ss[:, i].T
        axs[f'pred{i+1}'].fill_between(tstar, pred - std, pred + std, alpha=0.1)
        axs[f'pred{i+1}'].plot(tstar, pred)
        axs[f'pred{i+1}'].grid(which='major', alpha=0.5)
        axs[f'pred{i+1}'].grid(which='minor', alpha=0.2)


        resid = gprn.y[i] - a[:, i]
        axs[f'resid{i+1}'].errorbar(gprn.time, resid, gprn.yerr[i], fmt='ok',
                                    ms=2)
        axs[f'resid{i+1}'].axhline(y=0.0, ls='--', color='k', alpha=0.2)
        axs[f'resid{i+1}'].set_title(f'std: {resid.std():.2f}', loc='right',
                                     fontsize=10)
        equal_y_axis(axs[f'resid{i+1}'])
        axs[f'resid{i+1}'].set_ylabel('residuals')

        try:
            means.append(gprn.means[i](tstar))
        except TypeError:
            means.append(np.zeros_like(tstar))
        # axs[f'pred{i+1}'].plot(tstar, means[-1], ls='--')

    for i in range(gprn.p):
        axs[f'd{i+1}'].set(xlabel='')
        axs[f'd{i+1}'].set_ylabel('weight', color='C0')
        axs[f'd{i+1}'].set_title('weight(s) and mean', loc='left', fontsize=10)
        # weight(s)
        for w in bb[1][i::gprn.p]:
            axs[f'd{i+1}'].plot(tstar, w, alpha=0.6)
        axs[f'd{i+1}'].tick_params(axis='y', labelcolor='C0')

        # mean
        ax2 = axs[f'd{i+1}'].twinx()
        ax2.plot(tstar, means[i], color='k', ls='--', alpha=0.6)
        ax2.tick_params(axis='y', labelcolor='k')
        ax2.set_ylabel('mean', color='k')


    # axs['d2'].set(xlabel='')
    # axs['d2'].set_ylabel('weight', color='C0')
    # axs['d2'].set_title('weight(s) and mean', loc='left', fontsize=10)
    # axs['d2'].plot(tstar, bb[1][1 * gprn.q:].T, color='C0', alpha=0.6)
    # axs['d2'].tick_params(axis='y', labelcolor='C0')

    # ax2 = axs['d2'].twinx()
    # ax2.plot(tstar, means[1], color='C1', ls='--', alpha=0.6)
    # ax2.tick_params(axis='y', labelcolor='C1')
    # ax2.set_ylabel('mean', color='C1')

    # node
    axs['node'].set(xlabel='')
    _s = '' if gprn.q == 1 else 's'
    axs['node'].set_title('node' + _s, loc='left', fontsize=10)
    axs['node'].plot(tstar, bb[0].T, '-')

    equal_y_axis(axs['node'])
    ax0 = axs['pred1']
    for ax in axs.values():
        ax.sharex(ax0)

    for i in range(gprn.p - 1):
        plt.setp(axs[f'pred{i+1}'].get_xticklabels(), visible=False)
        plt.setp(axs[f'resid{i+1}'].get_xticklabels(), visible=False)
        plt.setp(axs[f'd{i+1}'].get_xticklabels(), visible=False)
    try:
        plt.setp(axs[f'pred{i+2}'].get_xticklabels(), visible=False)
    except KeyError:
        pass

    for ax in axs.values():
        ax.axvspan(gprn.time[0], gprn.time[-1], color='k', alpha=0.05,
                   zorder=-1)

    fig.tight_layout()

    return fig, axs



import numpy as np
from ._utils import Array, _array_input

__all__ = [
    'Constant', 'MultiConstant', 'Linear', 'Parabola', 'Cubic', 'Sine',
]


class meanFunction():
    """ Base class for mean functions"""
    _parsize = 0

    def __init__(self, *pars):
        self.pars = np.array(pars, dtype=float)

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__,
                                 ", ".join(map(str, self.pars)))

    def get_parameters(self):
        return self.pars

    @_array_input
    def set_parameters(self, p):
        msg = f'too few parameters for mean {self.__class__.__name__}'
        assert len(p) >= self.pars.size, msg
        if len(p) > self.pars.size:
            p = list(p)
            self.pars = np.array(p[:self.pars.size], dtype=float)
            for _ in range(self.pars.size):
                p.pop(0)
            return np.array(p)
        else:
            self.pars = p

    def __add__(self, b):
        return Sum(self, b)

    def __radd__(self, b):
        return self.__add__(b)

    def __mul__(self, b):
        return Product(self, b)

    def __rmul__(self, b):
        return self.__mul__(b)


class Sum(meanFunction):
    """ Sum of two mean functions """
    def __init__(self, m1, m2):
        self.m1, self.m2 = m1, m2
        if m1.__class__ == m2.__class__:
            # if they are the same class, number the parameter names
            param_names = []
            for p in m1._param_names:
                param_names.append(f'{p}1')
            for p in m2._param_names:
                param_names.append(f'{p}2')
            self._param_names = tuple(param_names)
        else:
            self._param_names = tuple(
                list(m1._param_names) + list(m2._param_names))
        self._parsize = m1._parsize + m2._parsize
        self.pars = np.r_[self.m1.pars, self.m2.pars]

    @_array_input
    def set_parameters(self, p):
        msg = f'too few parameters for mean {self.__class__.__name__}'
        assert len(p) >= self.pars.size, msg
        if len(p) > self.pars.size:
            self.pars = np.array(p[:self.pars.size], dtype=float)
            p = self.m1.set_parameters(p)
            p = self.m2.set_parameters(p)
            return p
        else:
            self.pars = p
            p = self.m1.set_parameters(p)
            p = self.m2.set_parameters(p)

    def __repr__(self):
        return "{0} + {1}".format(self.m1, self.m2)

    @_array_input
    def __call__(self, t):
        return self.m1(t) + self.m2(t)


class Product(meanFunction):
    """ Product of two mean functions """
    def __init__(self, m1, m2):
        self.m1, self.m2 = m1, m2
        self._param_names = tuple(
            list(m1._param_names) + list(m2._param_names))
        self._parsize = m1._parsize + m2._parsize
        self.pars = np.r_[self.m1.pars, self.m2.pars]

    @_array_input
    def set_parameters(self, p):
        msg = f'too few parameters for mean {self.__class__.__name__}'
        assert len(p) >= self.pars.size, msg
        if len(p) > self.pars.size:
            self.pars = np.array(p[:self.pars.size], dtype=float)
            p = self.m1.set_parameters(p)
            p = self.m2.set_parameters(p)
            return p
        else:
            self.pars = p
            p = self.m1.set_parameters(p)
            p = self.m2.set_parameters(p)

    def __repr__(self):
        return "{0} * {1}".format(self.m1, self.m2)

    @_array_input
    def __call__(self, t):
        return self.m1(t) * self.m2(t)


class Constant(meanFunction):
    """
    A constant mean function

    Args:
        c: The constant value of the mean function
    """
    _param_names = 'c',
    _parsize = 1

    def __init__(self, c: float):
        super(Constant, self).__init__(c)

    @_array_input
    def __call__(self, t):
        return np.full(t.shape, self.pars[0])


class MultiConstant(meanFunction):
    """
    Contant mean function for multiple instruments

    Args:
        offsets: Values of the between-instrument offsets and the average
                 value of the last instrument: [off_1, off_2, ..., avg_n]  
                 Offsets are relative to the last instrument.
        obsid: Indices of observations corresponding to each instrument. These
               should be one-based: [1, 1, ..., 2, 2, 2, ..., 3]
        time: Observed times. Should be the same size as `obsid`.
    """
    _parsize = 0

    def __init__(self, offsets: np.ndarray, obsid: np.ndarray, time: np.ndarray):
        self.obsid = obsid
        self.time = time
        self._parsize = (np.ediff1d(obsid) == 1).sum() + 1
        self.ii = obsid.astype(int) - 1

        if isinstance(offsets, float):
            offsets = [offsets]

        msg = 'wrong number of parameters, '
        msg += f'expected {self._parsize} got {len(offsets)}'
        assert len(offsets) == self._parsize, msg

        super().__init__(*offsets)
        self._param_names = [f'off{i}' for i in range(1, self._parsize)]
        self._param_names += ['mean']

    def time_bins(self):
        _1 = self.time[np.ediff1d(self.obsid, 0, None) != 0]
        _2 = self.time[np.ediff1d(self.obsid, None, 0) != 0]
        offset_times = np.mean((_1, _2), axis=0)
        return np.sort(np.r_[self.time[0], offset_times])

    @_array_input
    def __call__(self, t):
        offsets, c = self.pars[:-1], self.pars[-1]
        offsets = np.pad(offsets, (0, 1))

        if t.size == self.time.size:
            ii = self.ii
        else:
            time_bins = self.time_bins()
            ii = np.digitize(t, time_bins) - 1

        m = np.full_like(t, c) + np.take(offsets, ii)
        return m


class Linear(meanFunction):
    """
    A linear mean function, using the mean time as reference  
      `m(t) = slope * (t - mean(t)) + intercept`

    Args:
        slope: The slope of the linear function
        intercept: The intercept, using mean(t) as reference
    """
    _param_names = ('slope', 'intercept')
    _parsize = 2

    def __init__(self, slope: float, intercept: float):
        super(Linear, self).__init__(slope, intercept)

    @_array_input
    def __call__(self, t):
        tmean = t.mean()
        return self.pars[0] * (t - tmean) + self.pars[1]


class Parabola(meanFunction):
    """
    A 2nd degree polynomial mean function  
      `m(t) = quad * t² + slope * t + intercept`

    Args:
        quad: The quadratic term
        slope: The linear term
        intercept: The intercept
    """
    _param_names = ('slope', 'intercept', 'quadratic')
    _parsize = 3

    def __init__(self, quad: float, slope: float, intercept: float):
        super(Parabola, self).__init__(quad, slope, intercept)

    @_array_input
    def __call__(self, t):
        return np.polyval(self.pars, t)


class Cubic(meanFunction):
    """
    A 3rd degree polynomial mean function  
      `m(t) = cub * t³ + quad * t² + slope * t + intercept`

    Args:
        cub: The cubic term
        quad: The quadratic term
        slope: The linear term
        intercept: The intercept
    """
    _param_names = ('cub', 'quad', 'slope', 'intercept')
    _parsize = 4

    def __init__(self, cub: float, quad: float, slope: float, intercept: float):
        super(Cubic, self).__init__(cub, quad, slope, intercept)

    @_array_input
    def __call__(self, t):
        return np.polyval(self.pars, t)


class Sine(meanFunction):
    """
    A sinusoidal mean function  
      `m(t) = amplitude * sin(2*pi*t/P + phase)`

    Args:
        amplitude: The amplitude of the sinusoidal function
        period: The period of the sinusoidal function
        phase: The phase of the sinusoidal function
    """
    _param_names = ('amplitude', 'period', 'phase')
    _parsize = 3

    def __init__(self, amplitude: float, period: float, phase: float):
        super(Sine, self).__init__(amplitude, period, phase)

    @_array_input
    def __call__(self, t):
        A, P, φ = self.pars
        return A * np.sin((2 * np.pi * t / P) + φ)


# class Keplerian(meanFunction):
#     """ 
#     Keplerian function

#     Error: The Keplerian mean function is not yet implemented
         
#     """
#     _param_names = ('P', 'K', 'e', 'w', 'Tp')
#     _parsize = 5

#     def __init__(self, P, K, e, w, Tp):
#         raise NotImplementedError
#         super(Keplerian, self).__init__(P, K, e, w, Tp)

#     @_array_input
#     def __call__(self, t):
#         P, K, e, w, Tp = self.pars
#         return keplerian(t, P, K, e, w, Tp, 0.0)


# class CubicSun(meanFunction):
#     """
#     A 3rd degree mean function
#     m(t) = (t - xshift)**(-3) + yshift
#     """
#     _parsize = 4

#     def __init__(self, xshift, yshift):
#         super(Cubic, self).__init__(xshift, yshift)
#         self.xshift = xshift
#         self.yshift = yshift

#     @_array_input
#     def __call__(self, t):
#         return (t - self.xshift)**(-3) + self.yshift




"""
Collection of useful functions
"""

from typing import Union
from functools import wraps
from random import shuffle

import numpy as np
import matplotlib.pyplot as plt
from jax.numpy import ndarray as jnp_ndarray

from scipy.stats import invgamma
from scipy.linalg import cho_solve, cho_factor
from scipy.optimize import minimize

Array = Union[np.ndarray, jnp_ndarray]


def _array_input(f):
    """ Decorator to provide the __call__ methods with an array """
    @wraps(f)
    def wrapped(self, t):
        t = np.atleast_1d(t)
        r = f(self, t)
        return r
    return wrapped



##### Semi amplitude calculation ##############################################
def semi_amplitude(period, Mplanet, Mstar, ecc):
    """
    Calculates the semi-amplitude (K) caused by a planet with a given
    period and mass Mplanet, around a star of mass Mstar, with a
    eccentricity ecc.

    Parameters
    ----------
    period: float
        Period in years
    Mplanet: float
        Planet's mass in Jupiter masses, tecnically is the M.sin i
    Mstar: float
        Star mass in Solar masses
    ecc: float
        Eccentricity between 0 and 1

    Returns
    -------
    float
        Semi-amplitude K
    """
    per = np.float(np.power(1/period, 1/3))
    Pmass = Mplanet / 1
    Smass = np.float(np.power(1/Mstar, 2/3))
    Ecc = 1 / np.sqrt(1 - ecc**2)
    return 28.435 * per * Pmass* Smass * Ecc


##### Keplerian function ######################################################
def keplerian(P=365, K=.1, e=0, w=np.pi, T=0, phi=None, gamma=0, t=None):
    """
    keplerian() simulates the radial velocity signal of a planet in a
    keplerian orbit around a star.

    Parameters
    ----------
    P: float
        Period in days
    K: float
        RV amplitude
    e: float
        Eccentricity
    w: float
        Longitude of the periastron
    T: float
        Zero phase
    phi: float
        Orbital phase
    gamma: float
        Constant system RV
    t: array
        Time of measurements

    Returns
    -------
    t: array
        Time of measurements
    RV: array
        RV signal generated
    """
    if t is  None:
        print()
        print('TEMPORAL ERROR, time is nowhere to be found')
        print()
    #mean anomaly
    if phi is None:
        mean_anom = [2*np.pi*(x1-T)/P  for x1 in t]
    else:
        T = t[0] - (P*phi)/(2.*np.pi)
        mean_anom = [2*np.pi*(x1-T)/P  for x1 in t]
    #eccentric anomaly -> E0=M + e*sin(M) + 0.5*(e**2)*sin(2*M)
    E0 = [x + e*np.sin(x)  + 0.5*(e**2)*np.sin(2*x) for x in mean_anom]
    #mean anomaly -> M0=E0 - e*sin(E0)
    M0 = [x - e*np.sin(x) for x in E0]
    i = 0
    while i < 1000:
        #[x + y for x, y in zip(first, second)]
        calc_aux = [x2 - y for x2, y in zip(mean_anom, M0)]
        E1 = [x3 + y/(1-e*np.cos(x3)) for x3, y in zip(E0, calc_aux)]
        M1 = [x4 - e*np.sin(x4) for x4 in E0]
        i += 1
        E0 = E1
        M0 = M1
    nu = [2*np.arctan(np.sqrt((1+e)/(1-e))*np.tan(x5/2)) for x5 in E0]
    RV = [gamma + K*(e*np.cos(w)+np.cos(w+x6)) for x6 in nu] #m/s
    return t, RV


##### Phase-folding function ##################################################
def phase_folding(t, y, yerr, period):
    """
    phase_folding() allows the phase folding (duh...) of a given data
    accordingly to a given period

    Parameters
    ----------
    t: array
        Time
    y: array
        Measurements
    yerr: array
        Measurement errors
    period: float
        Period to fold the data

    Returns
    -------
    phase: array
        Phase
    folded_y: array
        Sorted measurments according to the phase
    folded_yerr:array
        Sorted errors according to the phase
    """
    #divide the time by the period to convert to phase
    foldtimes = t / period
    #remove the whole number part of the phase
    foldtimes = foldtimes % 1
    if yerr is None:
        yerr = 0 * y
    #sort everything
    phase, folded_y, folded_yerr = zip(*sorted(zip(foldtimes, y, yerr)))
    return phase, folded_y, folded_yerr


##### truncated cauchy distribution ###########################################
def truncCauchy_rvs(loc=0, scale=1, a=-1, b=1, size=None):
    """
    Generate random samples from a truncated Cauchy distribution.

    Parameters
    ----------
    loc: int
        Location parameter of the distribution
    scale: int
        Scale parameter of the distribution
    a, b: int
        Interval [a, b] to which the distribution is to be limited

    Returns
    -------
    rvs: float
        rvs of the truncated Cauchy
    """
    ua = np.arctan((a - loc)/scale)/np.pi + 0.5
    ub = np.arctan((b - loc)/scale)/np.pi + 0.5
    U = np.random.uniform(ua, ub, size=size)
    rvs = loc + scale * np.tan(np.pi*(U - 0.5))
    return rvs


##### inverse gamma distribution ###############################################
f = lambda x, lims: \
    (np.array([invgamma(a=x[0], scale=x[1]).cdf(lims[0]) - 0.01,
               invgamma(a=x[0], scale=x[1]).sf(lims[1]) - 0.01])**2).sum()

def invGamma(lower, upper, x0=[1, 5], showit=False):
    """
    Arguments
    ---------
    lower, upper : float
        The upper and lower limits between which we want 98% of the probability
    x0 : list, length 2
        Initial guesses for the parameters of the inverse gamma (a and scale)
    showit : bool
        Make a plot
    """
    limits = [lower, upper]
    result = minimize(f, x0=x0, args=limits, method='L-BFGS-B',
                      bounds=[(0, None), (0, None)], tol=1e-10)
    a, b = result.x
    if showit:
        _, ax = plt.subplots(1, 1, constrained_layout=True)
        d = invgamma(a=a, scale=b)
        x = np.linspace(0.2*limits[0], 2*limits[1], 1000)
        ax.plot(x, d.pdf(x))
        ax.vlines(limits, 0, d.pdf(x).max())
        plt.show()
    return invgamma(a=a, scale=b)


##### log sum ##################################################################
def log_sum(log_summands):
    """ log sum operation """
    a = np.inf
    x = log_summands.copy()
    while a == np.inf or a == -np.inf or np.isnan(a):
        a = x[0] + np.log(1 + np.sum(np.exp(x[1:] - x[0])))
        shuffle(x)
    return a


##### multivariate normal ######################################################
def multivariate_normal(r, c, method='cholesky'):
    """
    Computes multivariate normal density for "residuals" vector r and
    covariance c.

    :param array r:
        1-D array of k dimensions.

    :param array c:
        2-D array or matrix of (k x k).

    :param string method:
        Method used to compute multivariate density.
        Possible values are:
            * "cholesky": uses the Cholesky decomposition of the covariance c,
              implemented in scipy.linalg.cho_factor and scipy.linalg.cho_solve.
            * "solve": uses the numpy.linalg functions solve() and slogdet().

    :return array: multivariate density at vector position r.
    """
    # Compute normalization factor used for all methods.
    kk = len(r) * np.log(2*np.pi)
    if method == 'cholesky':
        # Use Cholesky decomposition of covariance.
        cho, lower = cho_factor(c)
        alpha = cho_solve((cho, lower), r)
        return -0.5 * (kk + np.dot(r, alpha) + 2 * np.sum(np.log(np.diag(cho))))
    if method == 'solve':
        # Use slogdet and solve
        (_, d) = np.linalg.slogdet(c)
        alpha = np.linalg.solve(c, r)
        return -0.5 * (kk + np.dot(r, alpha) + d)


##### RMS ######################################################################
def rms(array):
    """ Root mean square of array
        Parameters
        ----------
        array: array
            Measurements
            
        Returns
        -------
        rms: float
            Root mean squared error
    """
    mu = np.average(array)
    rms = np.sqrt(np.sum((array - mu)**2) / array.size)
    return rms


def wrms(array, weights):
    """ Weighted root mean square of array, given weights 
        
        Parameters
        ----------
        array: array
            Measurements
        weights: array
            weights = 1 / errors**2
            To add jitter do 1 / (errors*2 + jitter**2)
            
        Returns
        -------
        rms: float
            Weighted root mean squared error
    """
    mu = np.average(array, weights=weights)
    rms = np.sqrt(np.sum(weights * (array - mu)**2) / np.sum(weights)) 
    return rms


def anderson_darling_test(r):
    from scipy.stats import anderson
    result = anderson(r)
    s = result.significance_level[result.statistic > result.critical_values]
    if s.size == 0:
        return result, f'A-D: {result.significance_level[-1]:.0f}%'
    else:
        return result, f'A-D: {s.max():.0f}%'




import numpy as np

from scipy.linalg import cholesky, LinAlgError, inv
from scipy.stats import multivariate_normal
from scipy.optimize import minimize

class inference(object):
    """ 
    Class to perform mean field variational inference for GPRNs. 
    See Nguyen & Bonilla (2013) for more information.
    
    Parameters
    ----------
    num_nodes: int
        Number of latent node functions f(x), called f hat in the article
    time: array
        Time coordinates
    k: int 
        Mixture of k isotropic gaussian distributions
    *args: arrays
        The actual data (or components), it needs be given in order of data1, 
        data1error, data2, data2error, etc...
    """ 
    def  __init__(self, num_nodes, time, k, *args):
        #number of node functions; f(x) in Wilson et al. (2012)
        self.num_nodes = num_nodes
        self.q = num_nodes
        #array of the time
        self.time = time 
        #number of observations, N in Wilson et al. (2012)
        self.N = self.time.size
        #mixture of k isotropic gaussian distributions
        self.k  = k
        #the data, it should be given as data1, data1error, data2, ...
        self.args = args 
        #number of outputs y(x); p in Wilson et al. (2012)
        self.p = int(len(self.args)/2)
        #total number of weights, we will have q*p weights in total
        self.qp =  self.q * self.p
        self.d = self.time.size * self.q *(self.p+1)
        #to organize the data we now join everything
        self.tt = np.tile(time, self.p) #"extended" time
        ys = []
        ystd = []
        yerrs = []
        for i,j  in enumerate(args):
            if i%2 == 0:
                ys.append(j)
                ystd.append(np.std(j))
            else:
                yerrs.append(j)
        self.ystd = np.array(ystd).reshape(self.p, 1)
        self.y = np.array(ys).reshape(self.p, self.N) #matrix p*N of outputs
        self.yerr = np.array(yerrs).reshape(self.p, self.N) #matrix p*N of errors
        self.yerr2 = self.yerr**2
        #check if the input was correct
        assert int((i+1)/2) == self.p, \
        'Given data and number of components dont match'
        
        
##### mean functions definition ###############################################
    def _mean(self, means, time=None):
        """
        Returns the values of the mean functions
        
        Parameters
        ----------
        
        Returns
        -------
        m: float
            Value of the mean
        """
        if time is None:
            N = self.time.size
            m = np.zeros_like(self.tt)
            for i, meanfun in enumerate(means):
                if meanfun is None:
                    continue
                else:
                    m[i*N : (i+1)*N] = meanfun(self.time)
        else:
            N = time.size
            tt = np.tile(time, self.p)
            m = np.zeros_like(tt)
            for i, meanfun in enumerate(means):
                if meanfun is None:
                    continue
                else:
                    m[i*N : (i+1)*N] = meanfun(time)
        return m
    
    
##### To create matrices and samples ###########################################
    def _kernelMatrix(self, kernel, time = None):
        """
        Returns the covariance matrix created by evaluating a given kernel 
        at inputs time
        
        Parameters
        ----------
        
        Returns
        -------
        K: array
            Matrix of a covariance function
        """
        r = time[:, None] - time[None, :]
        K = kernel(r) + 1e-6*np.diag(np.diag(np.ones_like(r)))
        K[np.abs(K)<1e-12] = 0.
        return K
    
    
    def _predictKMatrix(self, kernel, time):
        """
        To be used in predict_gp()
        
        Parameters
        ----------
        
        Returns
        -------
        K: array
            Matrix of a covariance function
        """
        if time.size == 1:
            r = time - self.time[None, :]
        else:
            r = time[:,None] - self.time[None,:]
        K = kernel(r) 
        return K
    
    
    def _u_to_fhatW(self, u):
        """
        Given an array of values, divides it in the corresponding nodes (f hat)
        and weights (w) parts
        
        Parameters
        ----------
        u: array
        
        Returns
        -------
        f: array
            Samples of the nodes
        w: array
            Samples of the weights
        """
        f = u[:self.q * self.N].reshape((1, self.q, self.N))
        w = u[self.q * self.N:].reshape((self.p, self.q, self.N))
        return f, w
    
    
    def _cholNugget(self, matrix, maximum=10):
        """
        Returns the cholesky decomposition to a given matrix, if it is not
        positive definite, a nugget is added to its diagonal.
        
        Parameters
        ----------
        matrix: array
            Matrix to decompose
        maximum: int
            Number of times a nugget is added.
        
        Returns
        -------
        L: array
            Matrix containing the Cholesky factor
        nugget: float
            Nugget added to the diagonal
        """
        nugget = 0 #our nugget starts as zero
        try:
            nugget += 1e-15
            L = cholesky(matrix, lower=True, overwrite_a=True)
            return L, nugget
        except LinAlgError:
            n = 0 #number of tries
            while n < maximum:
                try:
                    L = cholesky(matrix + nugget*np.identity(matrix.shape[0]),
                                 lower=True, overwrite_a=True)
                    return L, nugget
                except LinAlgError:
                    nugget *= 10.0
                finally:
                    n += 1
            raise LinAlgError("Not positive definite, even with nugget.")
            
            
    def sampleIt(self, latentFunc, time=None):
        """
        Returns samples from the kernel
        
        Parameters
        ----------
        latentFunc: func
            Covariance function
        time: array
            Time array
        
        Returns
        -------
        norm: array
            Sample of K 
        """
        #print(latentFunc)
        if time is None:
            time = self.time
        mean = np.zeros_like(time)
        K = self._kernelMatrix(latentFunc, time)
        normal = multivariate_normal(mean, K, allow_singular=True).rvs()
        return normal
    
    
##### Non-Parametric Variational Inference functions ###########################
    def ELBOcalc(self, nodes, weights, meanf, jitters, iterations = 10000):
        """
        Function to use to calculate the evidence lower bound
        
        Parameters
        ----------
        nodes: array
            Node functions 
        weights: array
            Weight function
        means: array
            Mean functions
        jitters: array
            Jitter terms
        iterations: int
            Number of iterations 

            
        Returns
        -------
        ELBO: array
            Value of the ELBO per iteration

        """
        #initial variational parameters
        mu = np.random.rand(self.d, self.k).T
        var = np.ones_like(np.random.rand(1, self.k).T) # why?
        muF, muW = [], []
        for k in range(self.k):
            m1, m2 = self._u_to_fhatW(mu[k, :])
            muF.append(m1)
            muW.append(m2)
        muF = np.array(muF)
        muW = np.array(muW)
        ELBO = self.ELBOaux(nodes, weights, meanf, jitters, mu, var)
        elboArray = np.array([ELBO]) #To add new elbo values inside
        iterNumber = 1
        while iterNumber < iterations:
            #ELBO = self.ELBOaux(nodes, weights, meanf, jitters, mu, var)
            ELBO, mu, var = self.updateMUandVAR(nodes, weights, meanf, jitters, 
                                                mu, var)
            elboArray = np.append(elboArray, ELBO)
            iterNumber += 1
            #Stoping criteria:
            if iterNumber > 5:
                means = np.mean(elboArray[-5:])
                criteria = np.abs(np.std(elboArray[-5:]) / means)
                if criteria < 1e-3 and criteria !=0:
                    return ELBO, mu, var
        print('Max iterations reached')
        return ELBO, mu, var
    
    
    def ELBOaux(self, nodes, weights, meanf, jitters, mu, var):
        """
        Evidence Lower bound to use in ELBOcalc()
        
        Parameters
        ----------
        nodes: array
            Node functions 
        weights: array
            Weight function
        means: array
            Mean functions
        jitters: array
            Jitter terms
        mu: array
            Variational means
        var: array
            Variational variances
            
        Returns
        -------
        ELBO: float
            Evidence lower bound
        new_mu: array
            New variational means
        new_var: array
            New variational variances
        """
        muF, muW = [], []
        for k in range(self.k):
            m1, m2 = self._u_to_fhatW(mu[k, :])
            muF.append(m1)
            muW.append(m2)
        muF = np.array(muF)
        muW = np.array(muW)
        #nodes and means
        Kf = np.array([self._kernelMatrix(i, self.time) for i in nodes])
        invKf = np.array([inv(i) for i in Kf])
        Lf = np.array([self._cholNugget(i)[0] for i in Kf])
        Kw = np.array([self._kernelMatrix(j, self.time) for j in weights])
        invKw = np.array([inv(j) for j in Kw])
        Lw = np.array([self._cholNugget(j)[0] for j in Kw])
        #Entropy
        Entropy = self._entropy(mu, var)
        #print('entropy:', Entropy)
        #Expected log-likelihood
        ExpLoglike = self._expectedLogLike(nodes, weights, meanf, jitters, 
                                           muF, muW, var)
        #print('expLL:', ExpLoglike)
        #Expected log prior
        ExpLogprior = self._expectedLogPrior(Kf, invKf, Lf, Kw, invKw, Lw, 
                                             muF, muW, var, jitters)
        #print('expLP:', ExpLogprior)
        # print('expLL+expLP:', np.sum(ExpLoglike + ExpLogprior)/self.k)
        ELBO = np.sum(ExpLoglike + ExpLogprior)/self.k -Entropy
        return ELBO
    
    
    def _entropy(self, mu, var):
        varmin = 1e-7
        beta = np.ones((self.k,1)) / self.k
        S0 = np.array(mu - np.mean(mu, axis=0)).T
        S = np.sum(S0*S0, axis=0) - 2*(S0.T @ S0)
        S = np.sum(S0*S0, axis=0) + S.T
        S[S<0] = 0 #numerical noise can cause it to negative
        var = var**2 + varmin
        s = (var[:,None] + var[None,:])
        logP = -0.5*S/s - 0.5*self.d*np.log(s)
        logP[logP<0] = 0 #numerical noise can cause it to negative 
        a = np.zeros((self.k,1))
        for i in range(self.k):
            a[i] = -np.log(self.k) + np.log(np.sum(np.exp(logP[0,i])))
        entropy = np.float(a.T @ beta)
        return entropy
    
    
    def _expectedLogLike(self, nodes, weights, means, jitters, muF, muW, var):
        new_y = np.concatenate(self.y) - self._mean(means, self.time)
        new_y = np.array(np.array_split(new_y, self.p)).T 
        jitt2 = np.array(jitters)**2 
        errs = 0
        for i in range(self.p):
            errs += jitt2[i] + self.yerr2[i]
        ### first term of equation 3.22
        Wblk = []
        for k in range(self.k):
            Wblk.append(np.squeeze(muW[k]))
        Wblk = np.array(Wblk)
        Fblk = []
        for k in range(self.k):
            Fblk.append(np.squeeze(muF[k]))
        Fblk = np.array(Fblk)
        Ymean = Wblk * Fblk
        Ydiff = (new_y.T - Ymean)**2/errs
        logl = -0.5 * np.sum(Ydiff, axis=1)
        ### second term of equation 3.22
        kvals = []
        for k in range(self.k):
            value= 0
            for q in range(self.q):
                for p in range(self.p):
                    value = (muF[k,:,q,:] * muF[k,:,q,:])/(jitt2[p]+self.yerr2[p,:])
                    value += (muW[k,p,:,:] * muW[k,p,:,:])/(jitt2[p]+self.yerr2[p,:])
                    value += var[k]**4 * self.q /(jitt2[p] + np.sum(self.yerr2[p,:]))
                    kvals.append(self.p*var[k]**2*np.sum(value))
        kvals = np.array(np.squeeze(kvals))
        logl += -0.5*kvals
        ### third term of equation 3.22
        value = 0
        for p in range(self.p):
            for n in range(self.N):
                value += np.log(2*np.pi*(jitt2[p] + self.yerr2[p,n]))
        logl += -0.5*value
        return logl
        
        
    def _expectedLogPrior(self, Kf, invKf, Lf, Kw, invKw, Lw, 
                          muF, muW, var, jitters):
        ### first term
        logKf = [2*np.sum(np.log(np.diag(i))) for i in Lf]
        logKw = [2*np.sum(np.log(np.diag(i))) for i in Lw]
        logprior = -0.5*np.sum(logKf) -0.5*np.sum(logKw)
        ### second term 
        sum_kj, sum_kw = [], []
        for k in range(self.k):
            for q in range(self.q):
                mKfm = muF[k,:,q,:] @invKf[q] @muF[k,:,q,:].T
                vartracef = var[k]**2 *np.trace(invKf[q])
                for p in range(self.p):
                    #almost certain this will fail for more than 1 node
                    mKwm = muW[k,p,q,:] @invKw[p] @muW[k,p,q,:].T
                    vartracew = var[k]**2 *np.trace(invKw[p])
            sum_kj.append(np.float(mKfm+vartracef))
            sum_kw.append(np.float(mKwm+vartracew))
        logprior += -0.5*np.array(sum_kj) -0.5*np.array(sum_kw)
        return logprior
    
    
    def updateMUandVAR(self, nodes, weights, meanf, jitters, mu, var):
        res1 = minimize(self._updateMU, x0 = mu, 
                       args = (nodes, weights, meanf, jitters, var), 
                       method='Nelder-Mead', 
                       options={'disp': False, 
                                'maxiter': 200})
        mu  = res1.x
        res2 = minimize(self._updateVAR, x0 = var, 
                        args = (nodes, weights, meanf, jitters, mu), 
                        method='Nelder-Mead', 
                        options={'disp': False, 
                                'maxiter': 200})
        var  = res2.x
        mu = mu.reshape(self.k, self.d)
        ELBO = self.ELBOaux(nodes, weights, meanf, jitters, mu, var)
        return ELBO, mu, var

    def _updateMU(self, mu, nodes, weights, meanf, jitters, var):
        mu = mu.reshape(self.k, self.d)
        e = -self.ELBOaux(nodes, weights, meanf, jitters, mu, var)
        return e

    def _updateVAR(self, var, nodes, weights, meanf, jitters, mu):
        mu = mu.reshape(self.k, self.d)
        e = -self.ELBOaux(nodes, weights, meanf, jitters, mu, var)
        return e
    
    def _squaredDistance(self, X):
        m, n = X.shape
        D = np.zeros((n,n))
        for i in range(n):
            for j in range(i+1, n):
                D[i,j] = np.linalg.norm(X[:,i] - X[:,j])**2
                D[j,i] = D[i,j]
        return D


### END



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from setuptools import setup


# Get the long description from the README file
from os import path
here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='gpyrn',
    version='1.0.1',
    description='Gaussian process regression networks for exoplanet detection',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='João Camacho',
    author_email='joao.camacho@astro.up.pt',
    license='MIT',
    url='https://github.com/iastro-pt/gpyrn',
    packages=['gpyrn'],
)



"""
Computation of the evidence using the method developed by Perrakis et al. (2014)
"""
import random
from math import sqrt, log
import numpy as np
import scipy.stats
from gprn import utils

##### Original functions taken from https://github.com/exord/bayev #############
def compute_perrakis_estimate(marginal_sample, lnlikefunc, lnpriorfunc,
                              nsamples=1000, lnlikeargs=(), lnpriorargs=(),
                              densityestimation='histogram', 
                              errorestimation=False, **kwargs):
    """
    Computes the Perrakis estimate of the bayesian evidence.
    The estimation is based on n marginal posterior samples
    (indexed by s, with s = 0, ..., n-1).
    
    Parameters
    ----------
    :param array marginal_sample:
        A sample from the parameter marginal posterior distribution.
        Dimensions are (n x k), where k is the number of parameters.
    :param callable lnlikefunc:
        Function to compute ln(likelihood) on the marginal samples.
    :param callable lnpriorfunc:
        Function to compute ln(prior density) on the marginal samples.
    :param nsamples:
        Number of samples to produce.
    :param tuple lnlikeargs:
        Extra arguments passed to the likelihood function.
    :param tuple lnpriorargs:
        Extra arguments passed to the lnprior function.
    :param str densityestimation:
        The method used to estimate theinitial_samples marginal posterior density of each
        model parameter ("normal", "kde", or "histogram").
        
    Other parameters
    ----------------
    :param kwargs:
        Additional arguments passed to estimate_density function.
    :return:
        
    References
    ----------
    Perrakis et al. (2014; arXiv:1311.0674)
    
    Returns
    -------
    Perrakis estimate of the evidence
    """
    print('Estimating evidence...')
    if errorestimation:
        initial_sample = marginal_sample
    marginal_sample = make_marginal_samples(marginal_sample, nsamples)
    if not isinstance(marginal_sample, np.ndarray):
        marginal_sample = np.array(marginal_sample)
    number_parameters = marginal_sample.shape[1]
    marginal_posterior_density = np.zeros(marginal_sample.shape)
    for parameter_index in range(number_parameters):
        x = marginal_sample[:, parameter_index]
        marginal_posterior_density[:, parameter_index] = \
            estimate_density(x, method=densityestimation, **kwargs)
    prod_marginal_densities = marginal_posterior_density.prod(axis=1)
    log_prior = lnpriorfunc(marginal_sample, *lnpriorargs)
    log_likelihood = lnlikefunc(marginal_sample, *lnlikeargs)
    cond = log_likelihood != 0
    log_summands = (log_likelihood[cond] + log_prior[cond] -
                    np.log(prod_marginal_densities[cond]))
    perr = log_sum(log_summands) - log(len(log_summands))
    #error estimation
    K = 10
    if errorestimation:
        batchSize = initial_sample.shape[0]//K
        meanErr = [_perrakis_error(initial_sample[0:batchSize, :],
                                   lnlikefunc, lnpriorfunc, nsamples=nsamples,
                                   densityestimation=densityestimation)]
        for i in range(K):
            meanErr.append(_perrakis_error(initial_sample[i*batchSize:(i+1)*batchSize, :],
                                           lnlikefunc, lnpriorfunc,
                                           nsamples=nsamples,
                                           densityestimation=densityestimation))
        stdErr = np.std(meanErr)
        meanErr = np.mean(meanErr)
        print(perr, stdErr)
        return perr, stdErr
    return perr


def _perrakis_error(marginal_samples, lnlikefunc, lnpriorfunc, nsamples=1000,
                    densityestimation='histogram', errorestimation=False):
    """ To use when estimating the error of the perrakis method """
    return compute_perrakis_estimate(marginal_samples, lnlikefunc, lnpriorfunc,
                                     nsamples=nsamples,
                                     densityestimation=densityestimation,
                                     errorestimation=errorestimation)


def _errorCalc(marginal_sample, lnlikefunc, lnpriorfunc, nsamples=300,
               densityestimation='histogram', **kwargs):
    print('Estimating evidence error...')
    marginal_sample = make_marginal_samples(marginal_sample, nsamples)
    if not isinstance(marginal_sample, np.ndarray):
        marginal_sample = np.array(marginal_sample)
    number_parameters = marginal_sample.shape[1]
    #Estimate marginal posterior density for each parameter.
    marginal_posterior_density = np.zeros(marginal_sample.shape)
    for parameter_index in range(number_parameters):
        #Extract samples for this parameter._perrakis_error(
        x = marginal_sample[:, parameter_index]
        #Estimate density with method "densityestimation".
        marginal_posterior_density[:, parameter_index] = \
            estimate_density(x, method=densityestimation, **kwargs)
    #Compute produt of marginal posterior densities for all parameters
    prod_marginal_densities = marginal_posterior_density.prod(axis=1)
    #Compute lnprior and likelihood in marginal sample.
    log_prior = lnpriorfunc(marginal_sample)
    log_likelihood = lnlikefunc(marginal_sample)
    #Mask values with zero likelihood (a problem in lnlike)
    cond = log_likelihood != 0
    log_summands = (log_likelihood[cond] + log_prior[cond] -
                    np.log(prod_marginal_densities[cond]))
    perr = log_sum(log_summands) - log(len(log_summands))
    return perr


def estimate_density(x, method='histogram', **kwargs):
    """
    Estimate probability density based on a sample. Return value of density at
    sample points.
    :param array_like x: sample.
    :param str method:
        Method used for the estimation. 'histogram' estimates the density based
        on a normalised histogram of nbins bins; 'kde' uses a 1D non-parametric
        gaussian kernel; 'normal approximates the distribution by a normal
        distribution.
    Additional parameters
    :param int nbins:
        Number of bins used in "histogram method".
    :return: density estimation at the sample points.
    """
    nbins = kwargs.pop('nbins', 100)
    if method == 'normal':
        #Approximate each parameter distribution by a normal.
        return scipy.stats.norm.pdf(x, loc=x.mean(), scale=sqrt(x.var()))
    if method == 'kde':
        #Approximate each parameter distribution using a gaussian kernel estimation
        return scipy.stats.gaussian_kde(x)(x)
    if method == 'histogram':
        #Approximate each parameter distribution based on the histogram
        density, bin_edges = np.histogram(x, nbins, density=True)
        #Find to which bin each element corresponds
        density_indexes = np.searchsorted(bin_edges, x, side='left')
        #Correct to avoid index zero from being assiged to last element
        density_indexes = np.where(density_indexes > 0, density_indexes,
                                   density_indexes + 1)
        return density[density_indexes - 1]


def make_marginal_samples(joint_samples, nsamples=None):
    """
    Reshuffles samples from joint distribution of k parameters to obtain samples
    from the _marginal_ distribution of each parameter.
    :param np.array joint_samples:
        Samples from the parameter joint distribution. Dimensions are (n x k),
        where k is the number of parameters.
    :param nsamples:
        Number of samples to produce. If 0, use number of joint samples.
    :type nsamples:
        int or None
    """
    if nsamples > len(joint_samples) or nsamples is None:
        nsamples = len(joint_samples)
    marginal_samples = joint_samples[-nsamples:, :].copy()
    number_parameters = marginal_samples.shape[-1]
    # Reshuffle joint posterior samples to obtain _marginal_ posterior samples
    for parameter_index in range(number_parameters):
        random.shuffle(marginal_samples[:, parameter_index])
    return marginal_samples


def log_sum(log_summands):
    """ log_sum operation """
    a = np.inf
    x = log_summands.copy()
    while a == np.inf or a == -np.inf or np.isnan(a):
        a = x[0] + np.log(1 + np.sum(np.exp(x[1:] - x[0])))
        random.shuffle(x)
    return a


def compute_harmonicmean(lnlike_post, posterior_sample=None, lnlikefunc=None,
                         lnlikeargs=(), **kwargs):
    """
    Computes the harmonic mean estimate of the marginal likelihood.
    The estimation is based on n posterior samples
    (indexed by s, with s = 0, marginal likelihood error..., n-1), but can be done directly if the
    log(likelihood) in this sample is passed.
    :param array lnlike_post:
        log(likelihood) computed over a posterior sample. 1-D array of length n.
        If an emply array is given, then compute from posterior sample.
    :param array posterior_sample:
        A sample from the parameter posterior distribution.
        Dimensions are (n x k), where k is the number of parameters. If None
        the computation is done using the log(likelihood) obtained from the
        posterior sample.
    :param callable lnlikefunc:
        Function to compute ln(likelihood) on the marginal samples.
    :param tuple lnlikeargs:
        Extra arguments passed to the likelihood function.
    Other parameters
    ----------------
    :param int size:
        Size of sample to use for computation. If none is given, use size of
        given array or posterior sample.
    References
    ----------
    Kass & Raftery (1995), JASA vol. 90, N. 430, pp. 773-795
    """
    if len(lnlike_post) == 0 and posterior_sample is not None:
        samplesize = kwargs.pop('size', len(posterior_sample))
        if samplesize < len(posterior_sample):
            posterior_subsample = np.random.choice(posterior_sample,
                                                   size=samplesize,
                                                   replace=False)
        else:
            posterior_subsample = posterior_sample.copy()
        #Compute log likelihood in posterior sample.
        log_likelihood = lnlikefunc(posterior_subsample, *lnlikeargs)
    elif len(lnlike_post) > 0:
        samplesize = kwargs.pop('size', len(lnlike_post))
        log_likelihood = np.random.choice(lnlike_post, size=samplesize,
                                          replace=False)
    hme = -log_sum(-log_likelihood) + log(len(log_likelihood))
    return hme


def run_hme_mc(log_likelihood, nmc, samplesize):
    """ Harmonic mean """
    hme = np.zeros(nmc)
    for i in range(nmc):
        hme[i] = compute_harmonicmean(log_likelihood, size=samplesize)
    return hme


def compute_cj_estimate(posterior_sample, lnlikefunc, lnpriorfunc,
                        param_post, nsamples, qprob=None, lnlikeargs=(),
                        lnpriorargs=(), lnlike_post=None, lnprior_post=None):
    """
    Computes the Chib & Jeliazkov estimate of the bayesian evidence.
    The estimation is based on an posterior sample with n elements
    (indexed by s, with s = 0, ..., n-1), and a sample from the proposal
    distribution used in MCMC (qprob) of size nsample. Note that if qprob is
    None, it is estimated as a multivariate Gaussian.
    :param array posterior_sample:
        A sample from the parameter posterior distribution. Dimensions are
        (n x k), where k is the number of parameters.
    :param callable lnlikefunc:
        Function to compute ln(likelihood) on the marginal samples.
    :param callable lnpriorfunc:
        Function to compute ln(prior density) on the marginal samples.
    :param array param_post:
        Posterior parameter sample used to obtained fixed point needed by the
        algorithm.
    :param int nsamples:
        Size of sample drawn from proposal distribution.
    :param object or None qprob:
        Proposal distribution function. If None, it will be estimated as a
        multivariate Gaussian. If not None, it must possess the methods pdf and
        rvs. See scipy.stats.rv_continuous.
    :param tuple lnlikeargs:
        Extra arguments passed to the likelihood function.
    :param tuple lnpriorargs:
        Extra arguments passed to the lnprior function.
    :param array lnlike_post:
        log(likelihood) computed over a posterior sample. 1-D array of length n.
    :param array lnprior_post:
        log(prior) computed over a posterior sample. 1-D array of length n.
    :raises AttributeError:
        if instace qprob does not have method 'pdf' or 'rvs'.
    :raises TypeError:
        if methods 'pdf' or 'rvs' from instance qprob are not callable.
    :returns: Natural logarithm of estimated Bayesian evidence.
    References
    ----------
    Chib & Jeliazkov (2001): Journal of the Am. Stat. Assoc.; Mar 2001; 96, 453
    """
    #Find fixed point on which to estimate posterior ordinate.
    if lnlike_post is not None:
        #Pass values of log(likelihood) in posterior sample.
        arg_fp = [lnlike_post, ]
    else:
        #Pass function that computes log(likelihood).
        arg_fp = [lnlikefunc, ]
    if lnlike_post is not None:
        #Pass values of log(prior) in posterior sample.
        arg_fp.append(lnprior_post)
    else:
        #Pass function that computes log(prior).
        arg_fp.append(lnpriorfunc)
    fp, lnpost0 = get_fixed_point(posterior_sample, param_post, lnlikefunc, lnpriorfunc,
                                  lnlikeargs=lnlikeargs,
                                  lnpriorargs=lnpriorargs)
    #If proposal distribution is not given, define as multivariate Gaussian.
    if qprob is None:
        #Get covariance from posterior sample
        k = np.cov(posterior_sample.T)
        qprob = utils.MultivariateGaussian(fp, k)
    else:
        #Check that qprob has the necessary attributes
        for method in ('pdf', 'rvs'):
            try:
                att = getattr(qprob, method)
            except AttributeError:
                raise AttributeError('qprob does not have method '
                                     '\'{}\''.format(method))
            if not callable(att):
                raise TypeError('{} method of qprob is not '
                                'callable'.format(method))
    #Compute proposal density in posterior sample
    q_post = qprob.pdf(posterior_sample)
    #If likelihood over posterior sample is not given, compute it
    if lnlike_post is None:
        lnlike_post = lnlikefunc(posterior_sample, *lnlikeargs)
    #Idem for prior
    if lnprior_post is None:
        lnprior_post = lnpriorfunc(posterior_sample, *lnpriorargs)
    #Compute Metropolis ratio with respect to fixed point over posterior sample
    lnalpha_post = metropolis_ratio(lnprior_post + lnlike_post, lnpost0)
    #Sample from the proposal distribution with respect to fixed point
    proposal_sample = qprob.rvs(nsamples)
    #Compute likelihood and prior on proposal_sample
    lnprior_prop = lnpriorfunc(proposal_sample, *lnpriorargs)
    if np.all(lnprior_prop == -np.inf):
        raise ValueError('All samples from proposal density have zero prior'
                         'probability. Increase nsample.')
    #Now compute likelihood only on the samples where prior != 0.
    lnlike_prop = np.full_like(lnprior_prop, -np.inf)
    ind = lnprior_prop != -np.inf
    lnlike_prop[ind] = lnlikefunc(proposal_sample[ind, :], *lnlikeargs)
    #Get Metropolis ratio with respect to fixed point over proposal sample
    lnalpha_prop = metropolis_ratio(lnpost0, lnprior_prop + lnlike_prop)
    #Compute estimate of posterior ordinate (see Eq. 9 from reference)
    num = log_sum(lnalpha_post + q_post) - log(len(posterior_sample))
    den = log_sum(lnalpha_prop) - log(len(proposal_sample))
    lnpostord = num - den
    #Return log(Evidence) estimation
    return lnpost0 - lnpostord


def metropolis_ratio(lnpost0, lnpost1):
    """
    Compute Metropolis ratio for two states.
    :param float or array lnpost0:
        Value of ln(likelihood * prior) for inital state.
    :param float or array lnpost1:
        Value of ln(likelihood * prior) for proposal state.
    :raises ValueError: if lnpost0 and lnpost1 have different lengths.
    :return: log(Metropolis ratio)
    """
    if (hasattr(lnpost0, '__iter__') and hasattr(lnpost1, '__iter__') and
            len(lnpost0) != len(lnpost1)):
        raise ValueError('lnpost0 and lnpost1 have different lenghts.')
    return np.minimum(lnpost1 - lnpost0, 0.0)


def get_fixed_point(posterior_samples, param_post, lnlike, lnprior, lnlikeargs=(), lnpriorargs=()):
    """
    Find the posterior point closest to the model of the lnlike distribution.
    :param array posterior_samples:
        A sample from the parameters posterior distribution. Array dimensions
        must be (n x k), where n is the number of elements in the sample and
        k is the number of parameters.
    :param array or None param_post:
        A sample from the marginal posterior distribution of the parameter
        chosen to identify the high-density point to use as fixed point. This is
        typically one of the columns of posterior_samples, but could be any
        1-D array of size n. If None, then a multivariate Gaussian kernel
        estimate of the joint posterior distribution is used.
    :param array or callable lnlike:
        Function to compute log(likelihood). If an array is given, this is
        simply the log(likelihood) values at the posterior samples, and the
         best value will be chosen from this array.
    :param array or callable lnprior:
        Function to compute log(prior). If an array is given, this is
        simply the log(prior) values at the posterior samples, and the
        best value will be chosen from this array.
    :param tuple lnlikeargs:
        Extra arguments passed to lnlike functions.
    :param tuple lnpriorargs:
        Extra arguments passed to lnprior functions.
    :raises IndexError: if either lnlike or lnprior are arrays with length not
        matching the number of posterior samples.
    :return:
        the fixed point in parameter space and the value of
        log(prior * likelihood) evaluated at this point.
    """
    if param_post is not None:
        #Use median of param_post as fixed point.
        param0 = np.median(param_post)
        #Find argument closest to median.
        ind0 = np.argmin(np.abs(param_post - param0))
        fixed_point = posterior_samples[ind0, :]
        #Compute log(likelihood) at fixed_point
        if hasattr(lnlike, '__iter__'):
            if len(lnlike) != len(posterior_samples):
                raise IndexError('Number of elements in lnlike array and in '
                                 'posterior sample must match.')
            lnlike0 = lnlike[ind0]
        else:
            #Evaluate lnlike function at fixed point.
            lnlike0 = lnlike(fixed_point, *lnlikeargs)
        #Compute log(prior) at fixed_point
        if hasattr(lnprior, '__iter__'):
            if len(lnprior) != len(posterior_samples):
                raise IndexError('Number of elements in lnprior array and in '
                                 'posterior sample must match.')
            lnprior0 = lnprior[ind0]
        else:
            #Evaluate lnlike function at fixed point.
            lnprior0 = lnprior(fixed_point, *lnpriorargs)
        return fixed_point, lnlike0 + lnprior0
    raise NotImplementedError


### END



from functools import partial
from itertools import chain
import time as time_module
from typing import Tuple

import jax
import jax.numpy as jnp
from jax.scipy.linalg import cho_solve as cho_solve_jax
jax.config.update("jax_enable_x64", True)

from . import _gp, covfunc, meanfunc
from ._utils import Array, _array_input

import numpy as np
from scipy.linalg import cho_solve
from scipy.optimize import minimize
from scipy.stats import multivariate_normal
from emcee import EnsembleSampler, backends
from emcee.utils import sample_ellipsoid

# from numba import njit

from ._plots import plot_prediction

# def _cholNugget(matrix, maximum=1000):
#     """
#     Returns the cholesky decomposition to a given matrix, if it is not
#     positive definite, a nugget is added to its diagonal.

#     Parameters
#     ----------
#     matrix: array
#         Matrix to decompose
#     maximum: int
#         Number of times a nugget is added.

#     Returns
#     -------
#     L: array
#         Matrix containing the Cholesky factor
#     nugget: float
#         Nugget added to the diagonal
#     """
#     nugget = 0  # our nugget starts as zero
#     try:
#         nugget += 1e-15
#         L = cholesky(matrix, lower=True, overwrite_a=True)
#         return L, nugget
#     except LinAlgError:
#         n = 0 # number of tries
#         while n < maximum:
#             try:
#                 L = cholesky(matrix + nugget * np.identity(matrix.shape[0]),
#                              lower=True, overwrite_a=True)
#                 return L, nugget
#             except LinAlgError:
#                 nugget *= 10.0
#             finally:
#                 n += 1
#         raise LinAlgError("Not positive definite, even with nugget.")


def comp_results(a, b):
    if not np.allclose(a, b):
        print(a,b)
        raise Exception



# @njit
@jax.jit
def _cholNugget(matrix):
    """
    Returns the cholesky decomposition to a given matrix.

    Parameters
    ----------
    matrix: array
        Matrix to decompose

    Returns
    -------
    L: array
        Matrix containing the Cholesky factor
    nugget: float
        Nugget added to the diagonal
    """
    return jnp.linalg.cholesky(matrix), 0.0
    #  + 1.25e-6 * np.eye(matrix.shape[0])), 1.25e-6


class inference:
    """
    Mean-field variational inference for GPRNs.
    See Nguyen & Bonilla (2013) for more information.

    Args:
        q: int
            Number of latent node functions f(x)
        time: array
            Time coordinates
        *args: arrays
            The observed data in the following order:
                y1, y1error, y2, y2error, ...
    """
    def __init__(self, q: int, time: Array, *args):
        self.q = q
        self.time = time
        self.N = self.time.size

        # check if the input was correct
        msg = 'Number of observed data arrays should be even: y1, y1error, ...'
        assert len(args) > 0 and len(args) % 2 == 0, msg
        msg = 'Output arrays should all have the same dimensions as time'
        assert np.all(np.array([len(a) for a in args]) == self.N), msg

        # number of outputs
        self.p = int(len(args) / 2)
        # total number of weights
        self.qp = self.q * self.p
        self.d = self.N * self.q * (self.p + 1)

        # to organize the data, we now join everything
        self.tt = np.tile(time, self.p) # "extended" time
        self.y = np.concatenate([args[::2]])
        self.yerr = np.concatenate([args[1::2]])
        self.yerr2 = self.yerr**2

        self._components_set = False
        self._frozen_mask = np.array([])
        self._mu, self._var = None, None
        self._mu_var_iters = 0
        self.update_muvar_after = 50
        self.elbo_max_iter = 5000

    def set_components(self, nodes, weights, means, jitters):
        """
        Set the different GPRN components: nodes, weights, means, and jitters

        Args:
            nodes: `covFunction` or list of `covFunction`
                The q GPRN nodes. The number should match what was provided when
                creating this `inference`
            weights: `covFunction` or list of `covFunction`
                The q x p GPRN weights. The number should match the number of
                nodes times the number of datasets
            means: `meanFunction` or list of `meanFunction`
                The p GPRN mean functions
            jitters: float or list
                The p jitter values
        """
        if isinstance(nodes, covfunc.covFunction):
            nodes = [nodes]
        # check number
        if len(nodes) != self.q:
            msg = 'Wrong number of nodes provided, '
            msg += f'expected {self.q} got {len(nodes)}'
            raise ValueError(msg)

        if isinstance(weights, covfunc.covFunction):
            weights = [weights]
        # check number
        if len(weights) != self.qp:
            msg = 'Wrong number of weights provided, '
            msg += f'expected {self.qp} got {len(weights)}'
            raise ValueError(msg)

        if isinstance(means, (int, float, meanfunc.meanFunction)):
            means = [means]

        if isinstance(jitters, (int, float)):
            jitters = [jitters]

        self.nodes = nodes
        self.weights = weights
        self.means = means
        self.jitters = np.array(jitters, dtype=np.float)
        self._components_set = True

    def get_parameters(self, nodes=None, weights=None, means=None,
                       jitters=None, include_frozen=False):
        """
        Get the values of all the GPRN parameters
        """
        nones = [
            nodes is None, weights is None, means is None, jitters is None
        ]
        if not self._components_set and all(nones):
            msg = 'Cannot get parameters. '
            msg += 'Provide arguments or run set_components before.'
            raise ValueError(msg)

        if self._components_set:
            p = []
            for node in self.nodes:
                p.append(node.get_parameters())
            for weight in self.weights:
                p.append(weight.get_parameters())
            for mean in self.means:
                p.append(mean.get_parameters())
            for jitter in self.jitters:
                p.append(np.array([jitter]))
        else:
            p = []
            if nodes is not None:
                for node in nodes:
                    p.append(node.get_parameters())
            if weights is not None:
                for weight in weights:
                    p.append(weight.get_parameters())
            if means is not None:
                for mean in means:
                    p.append(mean.get_parameters())
            if jitters is not None:
                for jitter in jitters:
                    p.append(np.array([jitter]))

        if include_frozen:
            return np.concatenate(p).ravel()
        else:
            return np.concatenate(p).ravel()[~self.frozen_mask]

    @_array_input
    def set_parameters(self, parameters: Array):
        """
        Set values for all the GPRN parameters
        """
        msg = 'GPRN components not set, use set_components'
        assert self._components_set, msg
        all_parameters = self.get_parameters(include_frozen=True)
        n_free_parameters = self.n_parameters - self.frozen_mask.sum()

        if parameters.size == self.n_parameters:
            # all parameters provided, even if some may be frozen
            # we ignore the frozen ones
            parameters[self.frozen_mask] = all_parameters[self.frozen_mask]

        elif parameters.size == n_free_parameters:
            # only non-frozen parameters were provided, fill in the frozen ones
            for i, par in enumerate(all_parameters):
                if self.frozen_mask[i]:
                    parameters = np.insert(parameters, i, par)

        else:
            # wrong numer of parameters provided
            NP = parameters.size
            ep = self.n_parameters
            fp = n_free_parameters
            msg = f'Wrong number of parameters provided: got {NP}, '
            if ep == fp:
                msg += f'expected {ep}'
            else:
                msg += f'expected {ep} (all) or {fp} (not frozen)'
            raise ValueError(msg)

        it = [self.nodes, self.weights, self.means]
        for component in chain.from_iterable(it):
            parameters = component.set_parameters(parameters)
        self.jitters = parameters

    @property
    def n_parameters(self):
        """ Total number of parameters """
        msg = 'GPRN components not set, use set_components'
        assert self._components_set, msg
        n = 0
        it = [self.nodes, self.weights, self.means]
        for component in chain.from_iterable(it):
            n += component.pars.size
        n += self.jitters.size
        return n

    @property
    def parameters_dict(self):
        """ Dictionary with parameters names and values """
        msg = 'GPRN components not set, use set_components'
        assert self._components_set, msg

        p = {}
        for i, node in enumerate(self.nodes, start=1):
            for par, val in zip(node._param_names, node.pars):
                p[f'node{i}.{par}'] = val
        for i, weight in enumerate(self.weights, start=1):
            for par, val in zip(weight._param_names, weight.pars):
                p[f'weight{i}.{par}'] = val
        for i, mean in enumerate(self.means, start=1):
            for par, val in zip(mean._param_names, mean.pars):
                p[f'mean{i}.{par}'] = val
        for i, jit in enumerate(self.jitters, start=1):
            p[f'jitter{i}'] = jit
        return p

    def freeze_parameter(self, index=None, name=None):
        """
        Freeze (do not fit for) a given parameter, by index or name.

        Args:
            index: int, array
                Index (or indices) of the parameters to freeze
            name: str
                Name of the parameter to freeze. Including a "*" in `name`
                allows to freeze more than one parameter at once. For example,
                name='node1*' will freeze all parameters of the first node.
        """
        self.frozen_mask
        if index is None and name is None:
            raise ValueError('Provide either index or name')
        if name is None:
            self._frozen_mask[index] = True
        elif index is None:
            if '*' in name:
                names = list(self.parameters_dict.keys())
                name = name.replace('*', '')
                for index, known_name in enumerate(names):
                    if name in known_name:
                        self._frozen_mask[index] = True
            else:
                msg = f'Name "{name}" not found in parameters_dict'
                assert name in self.parameters_dict, msg
                index = list(self.parameters_dict.keys()).index(name)
                self._frozen_mask[index] = True

    def freeze_all_parameters(self):
        """ Freeze (do not fit for) all parameters """
        self._frozen_mask = np.ones(self._frozen_mask.size, dtype=bool)

    fix_parameter = freeze_parameter
    fix_all_parameters = freeze_all_parameters

    def thaw_parameter(self, index=None, name=None):
        """
        Thaw (free) a given parameter, by index or name.

        Args:
            index: int, array
                Index (or indices) of the parameters to thaw
            name: str
                Name of the parameter to thaw. Including a "*" in `name`
                allows to freeze more than one parameter at once. For example,
                name='node1*' will thaw all parameters of the first node.
        """
        self.frozen_mask
        if index is None and name is None:
            raise ValueError('Provide either index or name')
        if name is None:
            self._frozen_mask[index] = False
        elif index is None:
            if '*' in name:
                names = list(self.parameters_dict.keys())
                name = name.replace('*', '')
                for index, known_name in enumerate(names):
                    if name in known_name:
                        self._frozen_mask[index] = False
            else:
                msg = f'Name "{name}" not found in parameters_dict'
                assert name in self.parameters_dict, msg
                index = list(self.parameters_dict.keys()).index(name)
                self._frozen_mask[index] = False

    def thaw_all_parameters(self):
        """ Thaw (free) all parameters """
        self._frozen_mask = np.zeros(self._frozen_mask.size, dtype=bool)

    free_parameter = thaw_parameter
    free_all_parameters = thaw_all_parameters

    @property
    def frozen_mask(self):
        """ Boolean mask for the frozen parameters """
        msg = 'GPRN components not set, use set_components'
        assert self._components_set, msg
        if self._frozen_mask.size == 0:
            self._frozen_mask = np.full(self.n_parameters, False, dtype=bool)
        return self._frozen_mask

    @frozen_mask.setter
    def frozen_mask(self, mask):
        msg = 'Do not set frozen_mask, use thaw_parameter/freeze_parameter'
        raise NotImplementedError(msg)


    def _mean(self, means, time=None):
        """
        Returns the values of the mean functions

        Args:
            means : list of instances of meanFunction
            time : array, optional

        Returns:
            m: array
                Mean function evaluated at `time` or `self.time`
        """
        if time is None:
            N = self.time.size
            m = np.zeros_like(self.tt)
            for i, meanfun in enumerate(means):
                if meanfun is None:
                    continue
                else:
                    m[i * N:(i + 1) * N] = meanfun(self.time)
        else:
            N = time.size
            tt = np.tile(time, self.p)
            m = np.zeros_like(tt)
            for i, meanfun in enumerate(means):
                if meanfun is None:
                    continue
                else:
                    m[i * N:(i + 1) * N] = meanfun(time)
        return m

    def _KMatrix(self, kernel, time=None):
        """
        Returns the covariance matrix created by evaluating a given kernel at
        inputs time. For stability, a 1e-6 nugget is added to the diagonal.

        Args:
            kernel : instance of covFunc
            time : array, optional
        
        Returns:
            K: array
                Matrix of a covariance function
        """
        r = time[:, None] - time[None, :]
        K = kernel(r) + 1e-6 * np.eye(time.size)
        return K

    def _tinyNuggetKMatrix(self, kernel, time=None):
        """
        To be used in Prediction(). Returns the covariance matrix created by
        evaluating a given kernel at inputs time with the tiniest stability
        nugget possible.

        Returns:
            K: array Matrix of a covariance function
        """
        r = time[:, None] - time[None, :]
        K = kernel(r) + 1.25e-12 * np.diag(np.diag(np.ones_like(r)))
        return K

    def _predictKMatrix(self, kernel, time):
        """
        To be used in Prediction()

        Args:
            kernel : instance of covfunc
            time : array, optional

        Returns:
            K: array
                Matrix of a covariance function
        """
        if time.size == 1:
            r = time - self.time[None, :]
        else:
            r = time[:, None] - self.time[None, :]
        return kernel(r)

    def _u_to_fhatW(self, u):
        """
        Transform from the concatenated u values to the corresponding nodes (f)
        and weights (w).

        Args:
            u: array

        Returns:
            f: array
                Samples of the nodes
            w: array
                Samples of the weights
        """
        f = u[:self.q * self.N].reshape((1, self.q, self.N))
        w = u[self.q * self.N:].reshape((self.p, self.q, self.N))
        return f, w

    def _initMuVar(self, nodes, weights, jitter):
        a1 = [n.pars[0] for n in nodes]
        a2 = [w.pars[0] for w in weights]
        mean1, mean2 = [], []
        var1, var2 = [], []
        for _, n in enumerate(a1):
            m = [
                np.sqrt(np.abs(j) * n / i) * np.sign(j)
                for i, j in zip(a2, self.y)
            ]
            mean1.append(np.mean(m, axis=0))
            mean2.append(
                [np.sqrt(np.abs(j) * i / n) for i, j in zip(a2, self.y)])

            var1.append([np.mean(jitter) * np.ones_like(self.time)])
            var2.append([jitt * np.ones_like(self.time) for jitt in jitter])

        mu = np.concatenate((mean1, mean2), axis=None)
        var = np.concatenate((var1, var2), axis=None)
        return mu, var

    def _randomMuVar(self):
        mu = np.random.randn(self.d, 1)
        var = np.random.rand(self.d, 1)
        return mu, var

    def _sample_from_gp(self, kernel, time=None):
        """
        Returns random samples from a given kernel

        Args:
            kernel: covFunction
            time: array
        """
        if time is None:
            time = self.time
        mean = np.zeros_like(time)
        K = self._tinyNuggetKMatrix(kernel, time)
        normal = multivariate_normal(mean, K, allow_singular=True).rvs()
        return normal

    def sample(self, time=None):
        nodes, weights, means, jitters = self._get_components()
        node_samples = np.array([self._sample_from_gp(node) for node in nodes])
        weight_samples = np.array(
            [self._sample_from_gp(weight) for weight in weights])
        print(node_samples.shape)
        print(weight_samples.shape)
        return node_samples, weight_samples

    def _get_components(self, nodes=None, weights=None, means=None,
                        jitters=None):
        # if nothing was given, componentes must be already set
        all_none = all([i is None for i in (nodes, weights, means, jitters)])
        msg = 'GPRN components not set, use set_components'
        if all_none and not self._components_set:
            raise ValueError(msg)

        nodes = self.nodes if nodes is None else nodes
        weights = self.weights if weights is None else weights
        means = self.means if means is None else means
        jitters = self.jitters if jitters is None else jitters
        return nodes, weights, means, jitters


    @property
    def ELBO(self):
        """ The evidence lower bound for the GPRN """
        return self.ELBOcalc()[0]

    def ELBOcalc(self, nodes=None, weights=None, means=None, jitters=None,
                 max_iter=None, mu=None, var=None):
        """
        Calculate the evidence lower bound

        Args:
            nodes: list of `covFunction` instances, optional
                Kernel(s) for the node(s)
            weights: list of `covFunction` instances, optional
                Kernel(s) for the weight(s)
            means: list of `meanFunction` instances, optional
                Mean functions
            jitters: list of floats, optional
                Jitter terms
            max_iter: int, default self.elbo_max_iter
                Maximum number of iterations allowed in ELBO calculation
            mu: array or str, optional
                Variational means or 'init', 'random', or 'previous'
            var: array or str, optional
                Variational variances or 'init', 'random', or 'previous'

        Returns:
            ELBO: array
                Value of the ELBO per iteration
            mu: array
                Optimized variational means
            var: array
                Optimized variational variance (diagonal of sigma)
            """
        # deal with inputs or get attributes
        nodes, weights, means, jitters = self._get_components(
            nodes, weights, means, jitters)

        # initial variational parameters
        if mu is None or var is None:
            mu = var = 'init'

        if mu == 'previous' or var == 'previous':
            # should_update = self._mu_var_iters > self.update_muvar_after
            # if should_update:
            # mu, var = self._randomMuVar()
            # self._mu_var_iters = 0
            if self._mu is not None:
                mu, var = self._mu, self._var
                # self._mu_var_iters += 1
            else:
                mu, var = self._initMuVar(nodes, weights, jitters)

        elif mu == 'random' and var == 'random':
            mu, var = self._randomMuVar()

        elif mu == 'init' and var == 'init':
            mu, var = self._initMuVar(nodes, weights, jitters)

        if max_iter is None:
            max_iter = 10000

        j2 = np.array(jitters)**2
        Kf = np.array([self._KMatrix(i, self.time) for i in nodes])
        Kw = np.array([self._KMatrix(j, self.time) for j in weights])
        Lf = np.array([_cholNugget(j)[0] for j in Kf])
        Lw = np.array([_cholNugget(j)[0] for j in Kw])
        y = np.concatenate(self.y) - self._mean(means)
        y = np.array(np.array_split(y, self.p))

        # To add new elbo values inside
        ELBO, *_ = self.ELBOaux(Kf, Kw, Lf, Lw, y, j2, mu, var)
        elboArray = np.array([ELBO])
        iterNumber = 0

        if max_iter is None:
            max_iter = self.elbo_max_iter

        while iterNumber < max_iter:
            # Optimize mu and var analytically
            ELBO, mu, var, _, _ = self.ELBOaux(Kf, Kw, Lf, Lw, y, j2, mu, var)
            elboArray = np.append(elboArray, ELBO)
            iterNumber += 1
            # Stoping criteria:
            if iterNumber > 3:
                means = np.mean(elboArray[-3:])
                criteria = np.abs(np.std(elboArray[-3:]) / means)
                if criteria < 1e-3 and criteria != 0:
                    self._mu = mu
                    self._var = var
                    return ELBO, mu, var, iterNumber

        print('\nMax iterations reached')
        return ELBO, mu, var, iterNumber

    def ELBOaux(self, Kf, Kw, Lf, Lw, y, jitt2, mu, var):
        """
        Evidence Lower bound to use in ELBOcalc()

        Args:
            Kf: array
                Covariance matrices of the node functions 
            Kw: array
                Covariance matrices of the weight function
            Lf: array
                Lower matrix calculated with Cholesky of Kf 
            Lw: array
                Lower matrix calculated with Cholesky of Kw
            y: array
                Measurements - means
            jitt2: array
                Squared jitter terms
            mu: array
                Variational means
            var: array
                Variational variances

        Returns:
            ELBO: float
                Evidence lower bound
            new_mu: array
                New variational means
            new_var: array
                New variational variances
        """
        #to separate the variational parameters between the nodes and weights
        muF, muW = self._u_to_fhatW(mu.flatten())
        varF, varW = self._u_to_fhatW(var.flatten())
        sigmaF, muF, sigmaW, muW = self._updateSigMu(Kf, Kw, Lf, Lw, y, jitt2,
                                                     muF, varF, muW, varW)

        # new mean and var for the nodes
        muF = muF.reshape(1, self.q, self.N)
        varF = np.zeros_like(varF)
        for i in range(self.q):
            varF[:, i, :] = np.diag(sigmaF[i, :, :])

        # new mean and var for the weights
        varW = np.zeros_like(varW)
        for j in range(self.q):
            for i in range(self.p):
                varW[i, j, :] = np.diag(sigmaW[j, i, :, :])

        new_mu = np.concatenate((muF, muW))
        new_var = np.concatenate((varF, varW))

        # Entropy
        Ent = self._entropy(sigmaF, sigmaW)
        # Expected log prior
        LogP = self._expectedLogPrior(Kf, Kw, Lf, Lw, sigmaF, muF, sigmaW, muW)
        # Expected log-likelihood
        LogL = self._expectedLogLike(y, jitt2, sigmaF, muF, sigmaW, muW)
        # Evidence Lower Bound
        ELBO = (LogL + LogP + Ent) / self.q
        return ELBO, new_mu, new_var, sigmaF, sigmaW

    # @partial(jax.jit, static_argnums=(0,))
    def _updateSigMu(self, Kf, Kw, Lf, Lw, y, jitt2, muF, varF, muW, varW):
        """
        Efficient closed-form updates fot variational parameters. This
        corresponds to eqs. 16, 17, 18, and 19 of Nguyen & Bonilla (2013)

        Args:
            Kf: array
                Covariance matrices of the node functions
            Kw: array
                Covariance matrices of the weight function
            y: array
                Measurements - means
            jitt2: array
                Squared jitter terms
            muF: array
                Initial variational mean of each node
            varF: array
                Initial variational variance of each node
            muW: array
                Initial variational mean of each weight
            varW: array
                Initial variational variance of each weight

        Returns:
            sigma_f: array
                Updated variational covariance of each node
            mu_f: array
                Updated variational mean of each node
            sigma_w: array
                Updated variational covariance of each weight
            mu_w: array
                Updated variational mean of each weight
        """

        compare_results = False

        Kw = Kw.reshape(self.q, self.p, self.N, self.N)
        Lw = Lw.reshape(self.q, self.p, self.N, self.N)
        muF = np.squeeze(muF)

        # creation of Sigma_fj and mu_fj

        sigma_f = np.empty((self.q, self.N, self.N))
        mu_f = np.empty((self.q, self.N))

        # shape: p x N
        variance = jitt2[:, None] + self.yerr2

        # muW  shape: p x q x N
        # varW shape: p x q x N
        # sum is over p, need to replicate variance over q axis
        #? precomputed for all nodes j (eq 20)
        diagonal_vector = np.sum((muW * muW + varW) / variance[:, None, :],
                                 axis=0)


        for j in range(self.q):
            # Woodbury matrix identity
            sigma_f[j] = Kf[j] - Kf[j] @ np.linalg.solve(np.diag((1 / diagonal_vector[j])) + Kf[j], Kf[j])
            # # Kf_invf = np.linalg.inv(Kf[j])
            # Kf_inv = cho_solve((Lf[j], True), np.eye(self.N))
            # # assert np.allclose(Kf_invf, Kf_inv)
            # #? einsum gives a writeable view
            # Kf_inv_diagonal = np.einsum('jj->j', Kf_inv)
            # Kf_inv_diagonal += diagonal_vector[j]
            # sigma_f[j] = np.linalg.inv(Kf_inv)

            # muW  shape: p x q x N
            # muF shape: q x N
            # sum is over q, except j
            # y shape: p x N
            # print(muW * muF)
            # print('a')
            # print(np.delete(muW * muF, j, axis=1))
            # print('b')
            residuals = y - np.sum(np.delete(muW * muF, j, axis=1), axis=1)
            # residuals shape: p x N --> p x q x N
            pred = np.sum(residuals * muW[:, j, :] / variance,
                          axis=0)
            mu_f[j] = sigma_f[j] @ pred

        if compare_results:
            sigma_f_og, mu_f_og = [], []
            for j in range(self.q):
                diagFj, auxCalc = 0, 0
                for i in range(self.p):
                    diagFj = diagFj + (muW[i,j,:]*muW[i,j,:]+varW[i,j,:]) / (jitt2[i] + self.yerr2[i,:])
                    sumNj = np.zeros(self.N)
                    for k in range(self.q):
                        if k != j:
                            sumNj += muW[i, k, :] * muF[k, :].reshape(self.N)
                    auxCalc = auxCalc + ((y[i, :] - sumNj)*muW[i,j,:]) / (jitt2[i] + self.yerr2[i, :])
                    # R = y[i, :] - sumNj
                    # print(R)
                    # print(RR[j][i])
                    # input()

                comp_results(diagFj, diagonal_vector[j])
                CovF = np.diag(1 / diagFj) + Kf[j]
                sigF = Kf[j] - Kf[j] @ np.linalg.solve(CovF, Kf[j])

                # print(auxCalc)
                # input()

                sigma_f_og.append(sigF)
                mu_f_og.append(sigF @ auxCalc)
                # muF = np.array(mu_f_og)
            sigma_f_og = np.array(sigma_f_og)
            mu_f_og = np.array(mu_f_og)
            comp_results(sigma_f, sigma_f_og)
            comp_results(mu_f, mu_f_og)

        # #creation of Sigma_wij and mu_wij
        sigma_w = np.empty((self.q, self.p, self.N, self.N))
        mu_w = np.empty((self.p, self.q, self.N))
        #! why mu_w is not q x p x N ???

        # Kw shape: q x p x N x N
        # sigma_f shape: q x N x N
        # mu_f shape: q x N
        # print(sigma_f.shape)
        # print(mu_f.shape)

        # print(np.einsum('ijj->ij', sigma_f).shape)
        #? this one is also a view but the + creates a copy
        diagonal_vector = mu_f * mu_f + np.einsum('ijj->ij', sigma_f)
        # print(diagonal_vector.shape)
        # print(variance.shape)
        # input()
        # diagonal_vector = diagonal_vector / variance

        # mu_w = np.empty((self.q, self.N))

        for j in range(self.q):
            residuals = y - np.sum(np.delete(mu_f * muW, j, axis=1), axis=1)

            for i in range(self.p):
                sigma_w[j, i] = Kw[j, i] - Kw[j, i] @ np.linalg.solve(np.diag((variance[i] / diagonal_vector[j])) + Kw[j,i], Kw[j,i])
                # # Kw_invf = np.linalg.inv(Kw[j, i])
                # Kw_inv = cho_solve((Lw[j, i], True), np.eye(self.N))
                # # print(Kw_invf, Kw_inv)
                # # assert np.allclose(Kw_invf, Kw_inv, rtol=0.01)

                # Kw_inv_diagonal = np.einsum('jj->j', Kw_inv)
                # Kw_inv_diagonal += diagonal_vector[j] / variance[i]
                # sigma_w[j, i] = np.linalg.inv(Kw_inv)

                #
                # sum is over q, except j
                # y shape: p x N
                # residuals shape: p x N --> p x q x N
                pred = residuals[i] * mu_f[j, :] / variance[i]
                mu_w[i, j] = sigma_w[j, i] @ pred


        if compare_results:
            sigma_w_og, mu_w_og = [], np.zeros_like(muW)
            for j in range(self.q):
                for i in range(self.p):
                    mu_fj = mu_f_og[j]
                    var_fj = np.diag(sigma_f_og[j])
                    Diag_ij = (mu_fj * mu_fj + var_fj) / (jitt2[i] + self.yerr2[i, :])
                    Kw_ij = Kw[j, i, :, :]
                    CovWij = np.diag(1 / Diag_ij) + Kw_ij
                    sigWij = Kw_ij - Kw_ij @ np.linalg.solve(CovWij, Kw_ij)
                    sigma_w_og.append(sigWij)
                    sumNj = np.zeros(self.N)
                    for k in range(self.q):
                        if k != j:
                            sumNj += mu_f_og[k].reshape(self.N) * np.array(muW[i, k, :])
                    auxCalc = ((y[i, :] - sumNj) * mu_f_og[j, :]) / (jitt2[i] + self.yerr2[i, :])
                    mu_w_og[i, j, :] = sigWij @ auxCalc

            sigma_w_og = np.array(sigma_w_og).reshape(self.q, self.p, self.N, self.N)
            # mu_w_og = np.array(muW)

            comp_results(sigma_w, sigma_w_og)
            comp_results(mu_w, mu_w_og)

        # input('all good!')
        return sigma_f, mu_f, sigma_w, mu_w

    @partial(jax.jit, static_argnums=(0,))
    def _expectedLogLike(self, y, jitt2, sigma_f, mu_f, sigma_w, mu_w):
        """
        Calculates the expected log-likelihood in mean-field inference,
        corresponds to eq.14 in Nguyen & Bonilla (2013)

        Args:
            y: array
                Measurements - means
            jitt2: array
                Squared jitter terms
            sigma_f: array
                Variational covariance for each node
            mu_f: array
                Variational mean for each node
            sigma_w: array
                Variational covariance for each weight
            mu_w: array
                Variational mean for each weight

        Returns:
            logl: float
                Expected log-likelihood
        """
        compare_results = False


        # shape: p x N
        variance = jitt2[:, None] + self.yerr2

        logl = -0.5 * jnp.sum(jnp.log(2 * jnp.pi * variance))

        if compare_results:
            logl_og = 0
            for p in range(self.p):
                for n in range(self.N):
                    logl_og += np.log(2 * np.pi * (jitt2[p] + self.yerr2[p, n]))
            logl_og *= -0.5
            comp_results(logl, logl_og)

        # mu_f shape: 1 x q x N  (T: )
        # mu_w shape: p x q x N

        # shape: p x N                Nxqxp   Nxq
        Ωnu = jnp.einsum('ijk,ij->ik', mu_w.T, mu_f[0].T).T
        resid = self.y - Ωnu
        term2 = -0.5 * jnp.sum(resid**2 / variance)
        logl += term2

        if compare_results:
            sumN = []
            for n in range(self.N):
                for p in range(self.p):
                    Ydiff = y[p, n] - mu_f[0, :, n] @ mu_w[p, :, n].T
                    bottom = jitt2[p] + self.yerr2[p, n]
                    sumN.append((Ydiff.T * Ydiff) / bottom)
            term2_og = -0.5 * np.sum(sumN)
            comp_results(term2, term2_og)
            logl_og += term2_og


        # mu_f shape: 1 x q x N
        # mu_w shape: p x q x N
        #
        # sigma_f shape: q x N x N
        # sigma_w shape: q x p x N x N

        sigma_f_diagonals = jnp.einsum('ijj->ij', sigma_f)
        sigma_w_diagonals = jnp.einsum('ijkk->ijk', sigma_w)

        value = 0.0
        for i in range(self.p):
            for j in range(self.q):
                _1 = sigma_f_diagonals[j].T @ (mu_w[i, j]**2 / variance[i])
                _2 = sigma_w_diagonals[j, i].T @ (mu_f[0, j]**2 / variance[i])
                _3 = sigma_f_diagonals[j].T @ (sigma_w_diagonals[j, i] / variance[i])
                value += (_1 + _2 + _3)
        logl += -0.5 * value


        if compare_results:
            value_og = 0
            for p in range(self.p):
                for q in range(self.q):
                    value_og += np.sum(
                        (np.diag(sigma_f[q, :, :]) * mu_w[p, q, :] * mu_w[p, q, :]
                         + np.diag(sigma_w[q, p, :, :]) * mu_f[:, q, :] * mu_f[:, q, :]
                         + np.diag(sigma_f[q, :, :]) * np.diag(sigma_w[q, p, :, :])) / (jitt2[p] + self.yerr2[p, :])
                         )
            comp_results(value, value_og)
            logl_og += -0.5 * value_og

            comp_results(logl, logl_og)

            # input('all good')
        return logl

    @partial(jax.jit, static_argnums=(0,))
    def _expectedLogPrior(self, Kf, Kw, Lf, Lw, sigma_f, mu_f, sigma_w, mu_w):
        """
        Calculates the expection of the log prior wrt q(f,w) in mean-field
        inference, corresponds to eq.15 in Nguyen & Bonilla (2013)

        Args:
            Kf: array
                Covariance matrices of the node functions 
            Kw: array
                Covariance matrices of the weight function
            sigma_f: array
                Variational covariance for each node
            mu_f: array
                Variational mean for each node
            sigma_w: array
                Variational covariance for each weight
            mu_w: array
                Variational mean for each weight

        Returns:
            logp: float
                Expected log prior value
        """
        compare_results = False

        #we have Q nodes -> j in the paper; we have P y(x)s -> i in the paper
        Kw = Kw.reshape(self.q, self.p, self.N, self.N)
        Lw = Lw.reshape(self.q, self.p, self.N, self.N)
        muW = mu_w.reshape(self.q, self.p, self.N)

        first_term = 0.0  #calculation of the first term of eq.15 of Nguyen & Bonilla (2013)
        second_term = 0.0  #calculation of the second term of eq.15 of Nguyen & Bonilla (2013)
        sumSigmaF = jnp.zeros_like(sigma_f[0])

        for j in range(self.q):
            Lfj = Lf[j]
            logKf = jnp.sum(jnp.log(jnp.diag(Lfj)))

            mu_reshaped = mu_f[0, j, :]
            muKmu = mu_reshaped.T @ cho_solve_jax((Lfj, True), mu_reshaped)

            if compare_results:
                muK = np.linalg.solve(Lfj, mu_f[:, j, :].reshape(self.N))
                muKmu_og = muK @ muK
                comp_results(muKmu, muKmu_og)

            sumSigmaF = sumSigmaF + sigma_f[j]

            trace = jnp.trace(cho_solve_jax((Lfj, True), sumSigmaF))

            if compare_results:
                trace_og = np.trace(np.linalg.solve(Kf[j], sumSigmaF))
                comp_results(trace, trace_og)

            first_term += -logKf - 0.5 * (muKmu + trace)

            for i in range(self.p):
                muKmu = muW[j, i].T @ cho_solve_jax((Lw[j, i, :, :], True), muW[j, i])
                trace = jnp.trace(cho_solve_jax((Lw[j,i,:,:], True), sigma_w[j,i,:,:]))


                if compare_results:
                    muK = np.linalg.solve(Lw[j,i,:,:], muW[j,i])
                    muKmu_og = muK @ muK
                    trace_og = np.trace(np.linalg.solve(Kw[j,i,:,:], sigma_w[j,i,:,:]))
                    comp_results(muKmu, muKmu_og)
                    comp_results(trace, trace_og)


                second_term += -jnp.sum(jnp.log(jnp.diag(Lw[j,i,:,:]))) - 0.5*(muKmu + trace)

        const = -0.5 * self.N * self.q * (self.p + 1) * jnp.log(2 * jnp.pi)
        logp = first_term + second_term + const

        return logp

    @partial(jax.jit, static_argnums=(0,))
    def _entropy(self, sigma_f, sigma_w):
        """
        Calculates the entropy in mean-field inference, corresponds to eq.14 
        in Nguyen & Bonilla (2013)

        Args:
            sigma_f: array
                Variational covariance for each node
            sigma_w: array
                Variational covariance for each weight
        
        Returns:
            entropy: float
                Final entropy value
        """
        entropy = 0.0
        for j in range(self.q):
            L1 = _cholNugget(sigma_f[j])[0]
            entropy += jnp.sum(jnp.log(jnp.diag(L1)))
            for i in range(self.p):
                L2 = _cholNugget(sigma_w[j, i, :, :])[0]
                entropy += jnp.sum(jnp.log(jnp.diag(L2)))
        const = 0.5 * self.q * (self.p + 1) * self.N * (1 + jnp.log(2 * jnp.pi))
        return entropy + const

    def nELBO(self, parameters, max_iter=None):
        """ Return the negative ELBO for given values of the parameters """
        msg = 'GPRN components not set, use set_components'
        assert self._components_set, msg
        self.set_parameters(parameters)

        start = time_module.time()
        elbo, _, _, _ = self.ELBOcalc(self.nodes, self.weights, self.means,
                                      self.jitters, max_iter=max_iter,
                                      mu='previous', var='previous')
        end = time_module.time()

        spaces = 20 * ' '
        print(f'ELBO={elbo:7.2f} (took {1e3*(end-start):5.2f} ms){spaces}',
              end='\r', flush=True)
        # print()
        return -elbo


    def optimize(self, vars=None, **kwargs):
        """
        Optimize (maximize) the ELBO. If provided, `vars` controls the
        parameters which are free during the optimization.

        Args:
            vars : str or list, optional
                If provided, this defines the parameters included in the
                optimization process. Options are
                    vars = 'parameter_name'
                        optimize *only* parameter_name, all others are fixed
                    vars = '-parameter_name'
                        optimize all parameters *except* parameter_name
                    vars = [list of parameter_names]
                        optimize parameter_names and hold the others fixed
            **kwargs : dict
                Keyword arguments passed directly to scipy.optimize.minimize
        """
        if vars is not None:
            if isinstance(vars, str):
                if '-' in vars:
                    vars = vars.replace('-', '')
                    self.thaw_parameter(name='*')  # thaw all
                    self.freeze_parameter(name=vars)  # freeze vars
                else:
                    self.freeze_parameter(name='*')  # freeze all
                    self.thaw_parameter(name=vars)  # thaw vars
            elif isinstance(vars, list):
                self.freeze_parameter(name='*')  # freeze all
                for var in vars:
                    self.thaw_parameter(name=var)  # except all vars
            else:
                msg = f'`vars` should be str or list, got {type(vars)}'
                raise ValueError(msg)

        kwargs.setdefault('method', 'Nelder-Mead')
        res = minimize(self.nELBO, self.get_parameters(), **kwargs)
        self.set_parameters(res.x)
        return res

    def mcmc(self, priors, p0=None, vars=None, niter=500, **kwargs):
        """
        Sample the posterior distribution for the GPRN parameters. If provided,
        `vars` controls the parameters which are free in the MCMC.

        Args:
            priors: dict
                Dictionary with prior distributions for all (free) parameters
            p0: array, optional
                Initial values for the parameters. If not provided, a random
                sample from the corresponding prior will be used.
            vars : str or list, optional
                If provided, this defines the parameters included in the
                sampling process. Options are
                    vars = 'parameter_name'
                        sample *only* parameter_name, all others are fixed
                    vars = '-parameter_name'
                        sample all parameters *except* parameter_name
                    vars = [list of parameter_names]
                        sample parameter_names and hold the others fixed
            niter: int
                Number of MCMC iterations 
            **kwargs : dict
                Keyword arguments passed directly to emcee.EnsembleSampler
        """
        msg = 'GPRN components not set, use set_components'
        assert self._components_set, msg

        if vars is not None:
            if isinstance(vars, str):
                if '-' in vars:
                    vars = vars.replace('-', '')
                    self.thaw_parameter(name='*')  # thaw all
                    self.freeze_parameter(name=vars)  # freeze vars
                else:
                    self.freeze_parameter(name='*')  # freeze all
                    self.thaw_parameter(name=vars)  # thaw vars
            elif isinstance(vars, list):
                self.freeze_parameter(name='*')  # freeze all
                for var in vars:
                    self.thaw_parameter(name=var)  # except all vars
            else:
                msg = f'`vars` should be str or list, got {type(vars)}'
                raise ValueError(msg)

        all_parameter_names = np.array(list(self.parameters_dict.keys()))
        free_parameter_names = all_parameter_names[~self.frozen_mask]

        def prior_rvs():
            rvs = []
            for name in free_parameter_names:
                rvs.append(priors[name].rvs())
            return np.array(rvs)

        def logprior(parameters):
            _logprior = 0.0
            for par, name in zip(parameters, free_parameter_names):
                _logprior += priors[name].logpdf(par)
            return _logprior

        def logposterior(parameters):
            _logprior = logprior(parameters)
            if np.isneginf(_logprior):
                return -np.inf, -np.inf
            elbo = - self.nELBO(parameters, max_iter=100)
            return _logprior + elbo, elbo

        ndim = len(free_parameter_names)
        nwalkers = 2 * ndim

        print(f'Setting up sampler (parameters: {ndim}, walkers: {nwalkers})')

        if p0 is None:
            p0 = []
            for i in range(nwalkers - len(p0)):
                p0.append(prior_rvs())
            p0 = np.array(p0)
        else:
            sigma = []
            for name in free_parameter_names:
                try:
                    sigma.append(priors[name].std())
                except TypeError:
                    sigma.append(priors[name].std)

            p0 = sample_ellipsoid(p0, np.diag(sigma) / 100, size=nwalkers)
            for i, p in enumerate(p0):
                if np.isneginf(logprior(p)):
                    p0[i] = prior_rvs()

        print('initial values for parameters are set')
        _start = time_module.time()
        _ = [logposterior(p) for p in p0]
        _end = time_module.time()
        print()
        print(f'evaluation for initial values took {_end - _start:.0f} sec')
        print('- adjust your expectations accordingly')

        # Set up the backend
        filename = "gprn.h5"
        be = backends.HDFBackend(filename)
        be.reset(nwalkers, ndim)

        sampler = EnsembleSampler(nwalkers, ndim, logposterior, backend=be)

        # track the average autocorrelation time estimate
        index = 0
        autocorr = np.empty(niter)
        old_tau = np.inf

        for sample in sampler.sample(p0, iterations=niter, progress=True):
            if sampler.iteration % 10 == 0:
                print(sample.log_prob.max())
            # check convergence every 100 steps
            if sampler.iteration % 10:
                continue

            # Compute the autocorrelation time so far
            # Using tol=0 means that we'll always get an estimate even
            # if it isn't trustworthy
            tau = sampler.get_autocorr_time(tol=0)
            autocorr[index] = np.mean(tau)
            index += 1

            # Check convergence
            converged = np.all(tau * 100 < sampler.iteration)
            converged &= np.all(np.abs(old_tau - tau) / tau < 0.01)
            if converged:
                print('MCMC converged!')
                break
            old_tau = tau

        return sampler


    def _Prediction(self, nodes=None, weights=None, means=None, jitters=None,
                   tstar=None, mu=None, var=None, separate=False):
        """
        Prediction for mean-field inference

        Args:
            nodes: array
                Node functions 
            weights: array
                Weight function
            means: array
                Mean functions
            jitters: array
                Jitter terms
            tstar: array
                Predictions time
            mu: array, optional
                Variational means, use self._mu if not provided
            var: array, optional
                Variational variances, use self._var if not provided
            separate: bool
                Whether to return predictive for nodes and weights separately

        Returns:
            predictive mean(s): array
                Predicted means
            predictive variance(s): array
                Predictive variances
        """
        if nodes is None:
            nodes = self.nodes
        if weights is None:
            weights = self.weights
        if means is None:
            means = self.means
        if jitters is None:
            jitters = self.jitters

        if tstar is None:
            tstar = self.time

        if mu is None and var is None:
            if self._mu is None and self._var is None:
                mu, var = self._initMuVar(nodes, weights, jitters)
            else:
                mu, var = self._mu, self._var

        muF, muW = self._u_to_fhatW(mu.flatten())
        varF, varW = self._u_to_fhatW(var.flatten())
        meanVal = self._mean(means, tstar)
        meanVal = np.array(np.array_split(meanVal, self.p))
        y = np.concatenate(self.y) - self._mean(means)
        y = np.array(np.array_split(y, self.p))
        weights = np.array(weights).reshape(self.q, self.p)
        jitt2 = np.array(jitters)**2
        nPred, nVar = [], []
        wPred, wVar = [], []
        for q in range(self.q):
            gpObj = _gp.GP(self.time, muF[:, q, :])
            n, nv = gpObj.prediction(nodes[q], tstar,
                                     muF[:, q, :].reshape(self.N),
                                     varF[:, q, :].reshape(self.N))
            nPred.append(n)
            nVar.append(nv)
            for p in range(self.p):
                gpObj = _gp.GP(self.time, muW[p, q, :])
                w, wv = gpObj.prediction(weights[q, p], tstar,
                                         muW[p, q, :].reshape(self.N),
                                         varW[p, q, :].reshape(self.N))
                wPred.append(w)
                wVar.append(wv)
        nPred, nVar = np.array(nPred), np.array(nVar)

        wPredd = np.array(wPred).reshape(self.q, self.p, tstar.size)
        wVarr = np.array(wVar).reshape(self.q, self.p, tstar.size)
        predictives = np.zeros((tstar.size, self.p))
        predictivesVar = np.zeros((tstar.size, self.p))
        for p in range(self.p):
            predictives[:, p] += meanVal[p]
            for q in range(self.q):
                predictives[:, p] += nPred[q] * wPredd[q, p]
                predictivesVar[:,p] += wPredd[q,p]*wPredd[q,p]*nVar[q]\
                                       +wVarr[q,p]*(nVar[q] +nPred[q]*nPred[q])\
                                       +jitt2[p]
        wPred, wVar = np.array(wPred), np.array(wVar)

        if separate:
            predictives = np.array(predictives)
            sepPredictives = np.array([nPred, wPred], dtype=object)
            return predictives, predictivesVar, sepPredictives
        return predictives, predictivesVar

    def predict(self, tstar=None, nn=1000):
        """
        Get the GPRN prediction

        Args:
            tstar: array, optional
                Times at which to get prediction. If not provided, uses a
                linspace around `self.time`
            nn: int, default 1000
                Number of points in prediction, if `tstar` is not provided
        """
        if tstar is None:
            mi, ma = self.time.min(), self.time.max()
            tptp = self.time.ptp()
            tstar = np.linspace(mi - 0.2 * tptp, ma + 0.2 * tptp, nn)

        # a, v = self._Prediction()
        aa, vv, bb = self._Prediction(tstar=tstar, separate=True)
        ss = np.sqrt(vv)
        return tstar, aa, ss, bb


    plot_prediction = plot_prediction

    def plot_structure(self):
        raise NotImplementedError
        msg = 'GPRN components not set, use set_components'
        assert self._components_set, msg

        import daft
        pgm = daft.PGM()

        # observed datasets as "plates"
        pgm.add_plate([1.5, 0.2, 2, 3.2], label=r"exposure $i$", shift=-0.1)
        pgm.add_plate([2, 0.5, 1, 1], label=r"pixel $j$", shift=-0.1)
        pgm.render()